# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import sys
from subprocess import check_call

import pytest
import torch

from tests.common import (
    EXAMPLES_DIR,
    requires_cuda,
    requires_funsor,
    requires_horovod,
    requires_lightning,
    xfail_param,
)

logger = logging.getLogger(__name__)
pytestmark = pytest.mark.stage("test_examples")


CPU_EXAMPLES = [
    "air/main.py --num-steps=1",
    "air/main.py --num-steps=1 --no-baseline",
    "baseball.py --num-samples=200 --warmup-steps=100 --num-chains=2",
    "lkj.py --n=50 --num-chains=1 --warmup-steps=100 --num-samples=200",
    "capture_recapture/cjs.py --num-steps=1 -m 1",
    "capture_recapture/cjs.py --num-steps=1 -m 2",
    "capture_recapture/cjs.py --num-steps=1 -m 3",
    "capture_recapture/cjs.py --num-steps=1 -m 4",
    "capture_recapture/cjs.py --num-steps=1 -m 5",
    "capture_recapture/cjs.py --num-steps=1 -m 1 --tmc --tmc-num-samples=2",
    "capture_recapture/cjs.py --num-steps=1 -m 2 --tmc --tmc-num-samples=2",
    "capture_recapture/cjs.py --num-steps=1 -m 3 --tmc --tmc-num-samples=2",
    "capture_recapture/cjs.py --num-steps=1 -m 4 --tmc --tmc-num-samples=2",
    "capture_recapture/cjs.py --num-steps=1 -m 5 --tmc --tmc-num-samples=2",
    "contrib/autoname/scoping_mixture.py --num-epochs=1",
    "contrib/autoname/mixture.py --num-epochs=1",
    "contrib/autoname/tree_data.py --num-epochs=1",
    "contrib/cevae/synthetic.py --num-epochs=1",
    "contrib/epidemiology/sir.py --nojit -np=128 -t=2 -w=2 -n=4 -d=20 -p=1000 -f 2",
    "contrib/epidemiology/sir.py --nojit -np=128 -t=2 -w=2 -n=4 -d=20 -p=1000 -f 2 -c=2",
    "contrib/epidemiology/sir.py --nojit -np=128 -t=2 -w=2 -n=4 -d=20 -p=1000 -f 2 -e=2",
    "contrib/epidemiology/sir.py --nojit -np=128 -t=2 -w=2 -n=4 -d=20 -p=1000 -f 2 -k=1",
    "contrib/epidemiology/sir.py --nojit -np=128 -t=2 -w=2 -n=4 -d=20 -p=1000 -f 2 -e=2 -k=1",
    "contrib/epidemiology/sir.py --nojit -np=128 -t=2 -w=2 -n=4 -d=20 -p=1000 -f 2 --haar",
    "contrib/epidemiology/sir.py --nojit -np=128 -t=2 -w=2 -n=4 -d=20 -p=1000 -f 2 -nb=4",
    "contrib/epidemiology/sir.py --nojit -np=128 -t=2 -w=2 -n=4 -d=20 -p=1000 -f 2 -hfm=3",
    "contrib/epidemiology/sir.py --nojit -np=128 -t=2 -w=2 -n=4 -d=20 -p=1000 -f 2 -a",
    "contrib/epidemiology/sir.py --nojit -np=128 -t=2 -w=2 -n=4 -d=20 -p=1000 -f 2 -o=0.2",
    "contrib/epidemiology/sir.py --nojit -np=128 -ss=2 -n=4 -d=20 -p=1000 -f 2 --svi",
    "contrib/epidemiology/regional.py --nojit -t=2 -w=2 -n=4 -r=3 -d=20 -p=1000 -f 2",
    "contrib/epidemiology/regional.py --nojit -t=2 -w=2 -n=4 -r=3 -d=20 -p=1000 -f 2 --haar",
    "contrib/epidemiology/regional.py --nojit -t=2 -w=2 -n=4 -r=3 -d=20 -p=1000 -f 2 -hfm=3",
    "contrib/epidemiology/regional.py --nojit -t=2 -w=2 -n=4 -r=3 -d=20 -p=1000 -f 2 -nb=4",
    "contrib/epidemiology/regional.py --nojit -ss=2 -n=4 -r=3 -d=20 -p=1000 -f 2 --svi",
    "contrib/forecast/bart.py --num-steps=2 --stride=99999",
    "contrib/gp/sv-dkl.py --epochs=1 --num-inducing=4 --batch-size=1000",
    "contrib/gp/sv-dkl.py --binary --epochs=1 --num-inducing=4 --batch-size=1000",
    "contrib/mue/FactorMuE.py --test --small --include-stop --no-plots --no-save",
    "contrib/mue/FactorMuE.py --test --small -ard -idfac --no-substitution-matrix --no-plots --no-save",
    "contrib/mue/ProfileHMM.py --test --small --no-plots --no-save",
    "contrib/mue/ProfileHMM.py --test --small --include-stop --no-plots --no-save",
    "contrib/oed/ab_test.py --num-vi-steps=10 --num-bo-steps=2",
    "contrib/timeseries/gp_models.py -m imgp --test --num-steps=2",
    "contrib/timeseries/gp_models.py -m lcmgp --test --num-steps=2",
    "dmm.py --num-epochs=1",
    "dmm.py --num-epochs=1 --tmcelbo --tmc-num-samples=2",
    "dmm.py --num-epochs=1 --num-iafs=1",
    "dmm.py --num-epochs=1 --tmc --tmc-num-samples=2",
    "dmm.py --num-epochs=1 --tmcelbo --tmc-num-samples=2",
    "eight_schools/mcmc.py --num-samples=500 --warmup-steps=100",
    "eight_schools/svi.py --num-epochs=1",
    "einsum.py",
    "hmm.py --num-steps=1 --truncate=10 --model=0",
    "hmm.py --num-steps=1 --truncate=10 --model=1",
    "hmm.py --num-steps=1 --truncate=10 --model=2",
    "hmm.py --num-steps=1 --truncate=10 --model=3",
    "hmm.py --num-steps=1 --truncate=10 --model=4",
    "hmm.py --num-steps=1 --truncate=10 --model=5",
    "hmm.py --num-steps=1 --truncate=10 --model=6",
    "hmm.py --num-steps=1 --truncate=10 --model=6 --raftery-parameterization",
    "hmm.py --num-steps=1 --truncate=10 --model=7",
    "hmm.py --num-steps=1 --truncate=10 --model=0 --tmc --tmc-num-samples=2",
    "hmm.py --num-steps=1 --truncate=10 --model=1 --tmc --tmc-num-samples=2",
    "hmm.py --num-steps=1 --truncate=10 --model=2 --tmc --tmc-num-samples=2",
    "hmm.py --num-steps=1 --truncate=10 --model=3 --tmc --tmc-num-samples=2",
    "hmm.py --num-steps=1 --truncate=10 --model=4 --tmc --tmc-num-samples=2",
    "hmm.py --num-steps=1 --truncate=10 --model=5 --tmc --tmc-num-samples=2",
    "hmm.py --num-steps=1 --truncate=10 --model=6 --tmc --tmc-num-samples=2",
    "inclined_plane.py --num-samples=1",
    "lda.py --num-steps=2 --num-words=100 --num-docs=100 --num-words-per-doc=8",
    "minipyro.py --backend=pyro",
    "minipyro.py",
    "mixed_hmm/experiment.py --timesteps=1",
    "neutra.py -n 10 --num-warmup 10 --num-samples 10",
    "rsa/generics.py --num-samples=10",
    "rsa/hyperbole.py --price=10000",
    "rsa/schelling.py --num-samples=10",
    "rsa/schelling_false.py --num-samples=10",
    "rsa/semantic_parsing.py --num-samples=10",
    "scanvi/scanvi.py --num-epochs 1 --dataset mock",
    "sir_hmc.py -t=2 -w=2 -n=4 -d=2 -m=1 --enum",
    "sir_hmc.py -t=2 -w=2 -n=4 -d=2 -p=10000 --sequential",
    "sir_hmc.py -t=2 -w=2 -n=4 -d=100 -p=10000 -f 2",
    "smcfilter.py --num-timesteps=3 --num-particles=10",
    "sparse_gamma_def.py --num-epochs=2 --eval-particles=2 --eval-frequency=1 --guide custom",
    "sparse_gamma_def.py --num-epochs=2 --eval-particles=2 --eval-frequency=1 --guide auto",
    "sparse_gamma_def.py --num-epochs=2 --eval-particles=2 --eval-frequency=1 --guide easy",
    "svi_torch.py --num-epochs=2 --size=400",
    "svi_horovod.py --num-epochs=2 --size=400 --no-horovod",
    pytest.param(
        "svi_lightning.py --max_epochs=2 --size=400 --accelerator cpu --devices 1",
        marks=[requires_lightning],
    ),
    "toy_mixture_model_discrete_enumeration.py  --num-steps=1",
    "sparse_regression.py --num-steps=100 --num-data=100 --num-dimensions 11",
    "vae/ss_vae_M2.py --num-epochs=1",
    "vae/ss_vae_M2.py --num-epochs=1 --aux-loss",
    "vae/ss_vae_M2.py --num-epochs=1 --enum-discrete=parallel",
    "vae/ss_vae_M2.py --num-epochs=1 --enum-discrete=sequential",
    "vae/vae.py --num-epochs=1",
    "vae/vae_comparison.py --num-epochs=1",
    "cvae/main.py --num-quadrant-inputs=1 --num-epochs=1",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=0 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=1 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=2 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=3 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=4 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=5 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --raftery-parameterization ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=1 --tmc --tmc-num-samples=2 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=2 --tmc --tmc-num-samples=2 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=3 --tmc --tmc-num-samples=2 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=4 --tmc --tmc-num-samples=2 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=5 --tmc --tmc-num-samples=2 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --tmc --tmc-num-samples=2 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --tmc --tmc-num-samples=2  -rp",
]

CUDA_EXAMPLES = [
    "air/main.py --num-steps=1 --cuda",
    "baseball.py --num-samples=200 --warmup-steps=100 --num-chains=2 --cuda",
    "contrib/cevae/synthetic.py --num-epochs=1 --cuda",
    "contrib/epidemiology/sir.py --nojit -t=2 -w=2 -n=4 -d=20 -p=1000 -f 2 --cuda",
    "contrib/epidemiology/sir.py --nojit -t=2 -w=2 -n=4 -d=20 -p=1000 -f 2 -nb=16 --cuda",
    "contrib/epidemiology/sir.py --nojit -t=2 -w=2 -n=4 -d=20 -p=1000 -f 2 --haar --cuda",
    "contrib/epidemiology/regional.py --nojit -t=2 -w=2 -n=4 -r=3 -d=20 -p=1000 -f 2 --cuda",
    "contrib/epidemiology/regional.py --nojit -t=2 -w=2 -n=4 -r=3 -d=20 -p=1000 -f 2 --haar --cuda",
    "contrib/gp/sv-dkl.py --epochs=1 --num-inducing=4 --cuda",
    "contrib/mue/FactorMuE.py --test --small --include-stop --no-plots --no-save --cuda --cpu-data --pin-mem",
    "contrib/mue/FactorMuE.py --test --small -ard -idfac --no-substitution-matrix --no-plots --no-save --cuda",
    "contrib/mue/ProfileHMM.py --test --small --no-plots --no-save --cuda --cpu-data --pin-mem",
    "contrib/mue/ProfileHMM.py --test --small --include-stop --no-plots --no-save --cuda",
    "lkj.py --n=50 --num-chains=1 --warmup-steps=100 --num-samples=200 --cuda",
    "dmm.py --num-epochs=1 --cuda",
    "dmm.py --num-epochs=1 --num-iafs=1 --cuda",
    "dmm.py --num-epochs=1 --tmc --tmc-num-samples=2 --cuda",
    "dmm.py --num-epochs=1 --tmcelbo --tmc-num-samples=2 --cuda",
    "einsum.py --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=0 --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=1 --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=2 --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=3 --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=4 --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=5 --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=6 --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=6 --cuda --raftery-parameterization",
    "hmm.py --num-steps=1 --truncate=10 --model=7 --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=0 --tmc --tmc-num-samples=2 --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=1 --tmc --tmc-num-samples=2 --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=2 --tmc --tmc-num-samples=2 --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=3 --tmc --tmc-num-samples=2 --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=4 --tmc --tmc-num-samples=2 --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=5 --tmc --tmc-num-samples=2 --cuda",
    "hmm.py --num-steps=1 --truncate=10 --model=6 --tmc --tmc-num-samples=2 --cuda",
    "scanvi/scanvi.py --num-epochs 1 --dataset mock --cuda",
    "sir_hmc.py -t=2 -w=2 -n=4 -d=2 -m=1 --enum --cuda",
    "sir_hmc.py -t=2 -w=2 -n=4 -d=2 -p=10000 --sequential --cuda",
    "sir_hmc.py -t=2 -w=2 -n=4 -d=100 -p=10000 --cuda",
    "svi_torch.py --num-epochs=2 --size=400 --cuda",
    "svi_horovod.py --num-epochs=2 --size=400 --cuda --no-horovod",
    pytest.param(
        "svi_lightning.py --max_epochs=2 --size=400 --accelerator gpu --devices 1",
        marks=[requires_lightning],
    ),
    "vae/vae.py --num-epochs=1 --cuda",
    "vae/ss_vae_M2.py --num-epochs=1 --cuda",
    "vae/ss_vae_M2.py --num-epochs=1 --aux-loss --cuda",
    "vae/ss_vae_M2.py --num-epochs=1 --enum-discrete=parallel --cuda",
    "vae/ss_vae_M2.py --num-epochs=1 --enum-discrete=sequential --cuda",
    "cvae/main.py --num-quadrant-inputs=1 --num-epochs=1 --cuda",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=0 --cuda",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=1 --cuda",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=2 --cuda",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=3 --cuda",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=4 --cuda",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=5 --cuda",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --cuda",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --cuda --raftery-parameterization ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=1 --cuda--tmc --tmc-num-samples=2 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=2 --cuda--tmc --tmc-num-samples=2 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=3 --cuda--tmc --tmc-num-samples=2 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=4 --cuda--tmc --tmc-num-samples=2 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=5 --cuda--tmc --tmc-num-samples=2 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --cuda--tmc --tmc-num-samples=2 ",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --cuda--tmc --tmc-num-samples=2  -rp",
]


def xfail_jit(*args, **kwargs):
    reason = kwargs.pop("reason", "not jittable")
    return pytest.param(
        *args,
        marks=[
            pytest.mark.xfail(reason=reason),
            pytest.mark.skipif("CI" in os.environ, reason="slow test"),
        ]
    )


JIT_EXAMPLES = [
    "air/main.py --num-steps=1 --jit",
    xfail_jit(
        "baseball.py --num-samples=200 --warmup-steps=100 --jit",
        reason="unreproducible RuntimeError on CI",
    ),
    "contrib/autoname/mixture.py --num-epochs=1 --jit",
    "contrib/cevae/synthetic.py --num-epochs=1 --jit",
    "contrib/epidemiology/sir.py --jit -np=128 -t=2 -w=2 -n=4 -d=20 -p=1000 -f 2",
    "contrib/epidemiology/sir.py --jit -np=128 -ss=2 -n=4 -d=20 -p=1000 -f 2 --svi",
    "contrib/epidemiology/regional.py --jit -t=2 -w=2 -n=4 -r=3 -d=20 -p=1000 -f 2",
    "contrib/epidemiology/regional.py --jit -ss=2 -n=4 -r=3 -d=20 -p=1000 -f 2 --svi",
    xfail_jit("contrib/gp/sv-dkl.py --epochs=1 --num-inducing=4 --jit"),
    "contrib/mue/FactorMuE.py --test --small --include-stop --no-plots --no-save --jit",
    "contrib/mue/FactorMuE.py --test --small -ard -idfac --no-substitution-matrix --no-plots --no-save --jit",
    "contrib/mue/ProfileHMM.py --test --small --no-plots --no-save --jit",
    "contrib/mue/ProfileHMM.py --test --small --include-stop --no-plots --no-save --jit",
    xfail_jit("dmm.py --num-epochs=1 --jit"),
    xfail_jit("dmm.py --num-epochs=1 --num-iafs=1 --jit"),
    "eight_schools/mcmc.py --num-samples=500 --warmup-steps=100 --jit",
    "eight_schools/svi.py --num-epochs=1 --jit",
    "hmm.py --num-steps=1 --truncate=10 --model=1 --jit",
    "hmm.py --num-steps=1 --truncate=10 --model=2 --jit",
    "hmm.py --num-steps=1 --truncate=10 --model=3 --jit",
    "hmm.py --num-steps=1 --truncate=10 --model=4 --jit",
    "hmm.py --num-steps=1 --truncate=10 --model=5 --jit",
    "hmm.py --num-steps=1 --truncate=10 --model=7 --jit",
    xfail_jit(
        "hmm.py --num-steps=1 --truncate=10 --model=1 --tmc --tmc-num-samples=2 --jit"
    ),
    xfail_jit(
        "hmm.py --num-steps=1 --truncate=10 --model=2 --tmc --tmc-num-samples=2 --jit"
    ),
    xfail_jit(
        "hmm.py --num-steps=1 --truncate=10 --model=3 --tmc --tmc-num-samples=2 --jit"
    ),
    xfail_jit(
        "hmm.py --num-steps=1 --truncate=10 --model=4 --tmc --tmc-num-samples=2 --jit"
    ),
    "lda.py --num-steps=2 --num-words=100 --num-docs=100 --num-words-per-doc=8 --jit",
    "minipyro.py --backend=pyro --jit",
    "minipyro.py --jit",
    "sir_hmc.py -t=2 -w=2 -n=4 -d=2 -m=1 --enum --jit",
    "sir_hmc.py -t=2 -w=2 -n=4 -d=2 -p=10000 --sequential --jit",
    xfail_jit("sir_hmc.py -t=2 -w=2 -n=4 -p=10000 --jit"),
    xfail_jit("vae/ss_vae_M2.py --num-epochs=1 --aux-loss --jit"),
    "vae/ss_vae_M2.py --num-epochs=1 --enum-discrete=parallel --jit",
    "vae/ss_vae_M2.py --num-epochs=1 --enum-discrete=sequential --jit",
    "vae/ss_vae_M2.py --num-epochs=1 --jit",
    "vae/vae.py --num-epochs=1 --jit",
    "vae/vae_comparison.py --num-epochs=1 --jit",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=1 --jit",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=2 --jit",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=3 --jit",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=4 --jit",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=5 --jit",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --jit",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --jit --raftery-parameterization ",
]

HOROVOD_EXAMPLES = [
    "svi_horovod.py --num-epochs=2 --size=400",
    pytest.param(
        "svi_horovod.py --num-epochs=2 --size=400 --cuda", marks=[requires_cuda]
    ),
]

FUNSOR_EXAMPLES = [
    xfail_param(
        "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=0 --funsor",
        reason="unreproducible recursion error on travis?",
    ),
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=1 --funsor",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=2 --funsor",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=3 --funsor",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=4 --funsor",
    xfail_param(
        "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=5 --funsor",
        reason="https://github.com/pyro-ppl/pyro/issues/3046",
        run=False,
    ),
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --funsor",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --raftery-parameterization --funsor",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --jit --funsor",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --jit --raftery-parameterization --funsor",
    xfail_param(
        "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=0 --tmc --tmc-num-samples=2 --funsor",
        reason="unreproducible recursion error on travis?",
    ),
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=1 --tmc --tmc-num-samples=2 --funsor",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=2 --tmc --tmc-num-samples=2 --funsor",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=3 --tmc --tmc-num-samples=2 --funsor",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=4 --tmc --tmc-num-samples=2 --funsor",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=5 --tmc --tmc-num-samples=2 --funsor",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --tmc --tmc-num-samples=2 --funsor",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --tmc --tmc-num-samples=2 --funsor -rp",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --jit --tmc --tmc-num-samples=2 --funsor",
    "contrib/funsor/hmm.py --num-steps=1 --truncate=10 --model=6 --jit --tmc --tmc-num-samples=2 --funsor -rp",
]


def test_coverage():
    cpu_tests = set(
        (e if isinstance(e, str) else e.values[0]).split()[0] for e in CPU_EXAMPLES
    )
    cuda_tests = set(
        (e if isinstance(e, str) else e.values[0]).split()[0] for e in CUDA_EXAMPLES
    )
    jit_tests = set(
        (e if isinstance(e, str) else e.values[0]).split()[0] for e in JIT_EXAMPLES
    )
    for root, dirs, files in os.walk(EXAMPLES_DIR):
        for basename in files:
            if not basename.endswith(".py"):
                continue
            path = os.path.join(root, basename)
            with open(path) as f:
                text = f.read()
            example = os.path.relpath(path, EXAMPLES_DIR)
            if "__main__" in text:
                if example not in cpu_tests:
                    pytest.fail(
                        "Example: {} not covered in CPU_EXAMPLES.".format(example)
                    )
                if "--cuda" in text and example not in cuda_tests:
                    pytest.fail(
                        "Example: {} not covered by CUDA_EXAMPLES.".format(example)
                    )
                if "--jit" in text and example not in jit_tests:
                    pytest.fail(
                        "Example: {} not covered by JIT_EXAMPLES.".format(example)
                    )


@pytest.mark.parametrize("example", CPU_EXAMPLES)
def test_cpu(example):
    logger.info("Running:\npython examples/{}".format(example))
    example = example.split()
    filename, args = example[0], example[1:]
    filename = os.path.join(EXAMPLES_DIR, filename)
    check_call([sys.executable, filename] + args)


@requires_cuda
@pytest.mark.parametrize("example", CUDA_EXAMPLES)
def test_cuda(example):
    logger.info("Running:\npython examples/{}".format(example))
    example = example.split()
    filename, args = example[0], example[1:]
    filename = os.path.join(EXAMPLES_DIR, filename)
    check_call([sys.executable, filename] + args)


@pytest.mark.parametrize("example", JIT_EXAMPLES)
def test_jit(example):
    logger.info("Running:\npython examples/{}".format(example))
    example = example.split()
    filename, args = example[0], example[1:]
    filename = os.path.join(EXAMPLES_DIR, filename)
    check_call([sys.executable, filename] + args)


@requires_horovod
@pytest.mark.parametrize("np", [1, 2])
@pytest.mark.parametrize("example", HOROVOD_EXAMPLES)
def test_horovod(np, example):
    if "cuda" in example and np > torch.cuda.device_count():
        pytest.skip()
    horovodrun = "horovodrun -np {} --mpi-args=--oversubscribe".format(np)
    logger.info("Running:\n{} python examples/{}".format(horovodrun, example))
    example = example.split()
    filename, args = example[0], example[1:]
    filename = os.path.join(EXAMPLES_DIR, filename)
    check_call(horovodrun.split() + [sys.executable, filename] + args)


@requires_funsor
@pytest.mark.parametrize("example", FUNSOR_EXAMPLES)
def test_funsor(example):
    logger.info("Running:\npython examples/{}".format(example))
    example = example.split()
    filename, args = example[0], example[1:]
    filename = os.path.join(EXAMPLES_DIR, filename)
    check_call([sys.executable, filename] + args)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import torch

from pyro.infer import TraceMeanField_ELBO
from pyro.infer.util import torch_backward, torch_item


def conditional(
    Xnew,
    X,
    kernel,
    f_loc,
    f_scale_tril=None,
    Lff=None,
    full_cov=False,
    whiten=False,
    jitter=1e-6,
):
    r"""
    Given :math:`X_{new}`, predicts loc and covariance matrix of the conditional
    multivariate normal distribution

    .. math:: p(f^*(X_{new}) \mid X, k, f_{loc}, f_{scale\_tril}).

    Here ``f_loc`` and ``f_scale_tril`` are variation parameters of the variational
    distribution

    .. math:: q(f \mid f_{loc}, f_{scale\_tril}) \sim p(f | X, y),

    where :math:`f` is the function value of the Gaussian Process given input :math:`X`

    .. math:: p(f(X)) \sim \mathcal{N}(0, k(X, X))

    and :math:`y` is computed from :math:`f` by some likelihood function
    :math:`p(y|f)`.

    In case ``f_scale_tril=None``, we consider :math:`f = f_{loc}` and computes

    .. math:: p(f^*(X_{new}) \mid X, k, f).

    In case ``f_scale_tril`` is not ``None``, we follow the derivation from reference
    [1]. For the case ``f_scale_tril=None``, we follow the popular reference [2].

    References:

    [1] `Sparse GPs: approximate the posterior, not the model
    <https://www.prowler.io/sparse-gps-approximate-the-posterior-not-the-model/>`_

    [2] `Gaussian Processes for Machine Learning`,
    Carl E. Rasmussen, Christopher K. I. Williams

    :param torch.Tensor Xnew: A new input data.
    :param torch.Tensor X: An input data to be conditioned on.
    :param ~pyro.contrib.gp.kernels.kernel.Kernel kernel: A Pyro kernel object.
    :param torch.Tensor f_loc: Mean of :math:`q(f)`. In case ``f_scale_tril=None``,
        :math:`f_{loc} = f`.
    :param torch.Tensor f_scale_tril: Lower triangular decomposition of covariance
        matrix of :math:`q(f)`'s .
    :param torch.Tensor Lff: Lower triangular decomposition of :math:`kernel(X, X)`
        (optional).
    :param bool full_cov: A flag to decide if we want to return full covariance
        matrix or just variance.
    :param bool whiten: A flag to tell if ``f_loc`` and ``f_scale_tril`` are
        already transformed by the inverse of ``Lff``.
    :param float jitter: A small positive term which is added into the diagonal part of
        a covariance matrix to help stablize its Cholesky decomposition.
    :returns: loc and covariance matrix (or variance) of :math:`p(f^*(X_{new}))`
    :rtype: tuple(torch.Tensor, torch.Tensor)
    """
    # p(f* | Xnew, X, kernel, f_loc, f_scale_tril) ~ N(f* | loc, cov)
    # Kff = Lff @ Lff.T
    # v = inv(Lff) @ f_loc  <- whitened f_loc
    # S = inv(Lff) @ f_scale_tril  <- whitened f_scale_tril
    # Denote:
    #     W = (inv(Lff) @ Kf*).T
    #     K = W @ S @ S.T @ W.T
    #     Q** = K*f @ inv(Kff) @ Kf* = W @ W.T
    # loc = K*f @ inv(Kff) @ f_loc = W @ v
    # Case 1: f_scale_tril = None
    #     cov = K** - K*f @ inv(Kff) @ Kf* = K** - Q**
    # Case 2: f_scale_tril != None
    #     cov = K** - Q** + K*f @ inv(Kff) @ f_cov @ inv(Kff) @ Kf*
    #         = K** - Q** + W @ S @ S.T @ W.T
    #         = K** - Q** + K

    N = X.size(0)
    M = Xnew.size(0)
    latent_shape = f_loc.shape[:-1]

    if Lff is None:
        Kff = kernel(X).contiguous()
        Kff.view(-1)[:: N + 1] += jitter  # add jitter to diagonal
        Lff = torch.linalg.cholesky(Kff)
    Kfs = kernel(X, Xnew)

    # convert f_loc_shape from latent_shape x N to N x latent_shape
    f_loc = f_loc.permute(-1, *range(len(latent_shape)))
    # convert f_loc to 2D tensor for packing
    f_loc_2D = f_loc.reshape(N, -1)
    if f_scale_tril is not None:
        # convert f_scale_tril_shape from latent_shape x N x N to N x N x latent_shape
        f_scale_tril = f_scale_tril.permute(-2, -1, *range(len(latent_shape)))
        # convert f_scale_tril to 2D tensor for packing
        f_scale_tril_2D = f_scale_tril.reshape(N, -1)

    if whiten:
        v_2D = f_loc_2D
        W = torch.linalg.solve_triangular(Lff, Kfs, upper=False).t()
        if f_scale_tril is not None:
            S_2D = f_scale_tril_2D
    else:
        pack = torch.cat((f_loc_2D, Kfs), dim=1)
        if f_scale_tril is not None:
            pack = torch.cat((pack, f_scale_tril_2D), dim=1)

        Lffinv_pack = torch.linalg.solve_triangular(Lff, pack, upper=False)
        # unpack
        v_2D = Lffinv_pack[:, : f_loc_2D.size(1)]
        W = Lffinv_pack[:, f_loc_2D.size(1) : f_loc_2D.size(1) + M].t()
        if f_scale_tril is not None:
            S_2D = Lffinv_pack[:, -f_scale_tril_2D.size(1) :]

    loc_shape = latent_shape + (M,)
    loc = W.matmul(v_2D).t().reshape(loc_shape)

    if full_cov:
        Kss = kernel(Xnew)
        Qss = W.matmul(W.t())
        cov = Kss - Qss
    else:
        Kssdiag = kernel(Xnew, diag=True)
        Qssdiag = W.pow(2).sum(dim=-1)
        # Theoretically, Kss - Qss is non-negative; but due to numerical
        # computation, that might not be the case in practice.
        var = (Kssdiag - Qssdiag).clamp(min=0)

    if f_scale_tril is not None:
        W_S_shape = (Xnew.size(0),) + f_scale_tril.shape[1:]
        W_S = W.matmul(S_2D).reshape(W_S_shape)
        # convert W_S_shape from M x N x latent_shape to latent_shape x M x N
        W_S = W_S.permute(list(range(2, W_S.dim())) + [0, 1])

        if full_cov:
            St_Wt = W_S.transpose(-2, -1)
            K = W_S.matmul(St_Wt)
            cov = cov + K
        else:
            Kdiag = W_S.pow(2).sum(dim=-1)
            var = var + Kdiag
    else:
        if full_cov:
            cov = cov.expand(latent_shape + (M, M))
        else:
            var = var.expand(latent_shape + (M,))

    return (loc, cov) if full_cov else (loc, var)


def train(gpmodule, optimizer=None, loss_fn=None, retain_graph=None, num_steps=1000):
    """
    A helper to optimize parameters for a GP module.

    :param ~pyro.contrib.gp.models.GPModel gpmodule: A GP module.
    :param ~torch.optim.Optimizer optimizer: A PyTorch optimizer instance.
        By default, we use Adam with ``lr=0.01``.
    :param callable loss_fn: A loss function which takes inputs are
        ``gpmodule.model``, ``gpmodule.guide``, and returns ELBO loss.
        By default, ``loss_fn=TraceMeanField_ELBO().differentiable_loss``.
    :param bool retain_graph: An optional flag of ``torch.autograd.backward``.
    :param int num_steps: Number of steps to run SVI.
    :returns: a list of losses during the training procedure
    :rtype: list
    """
    optimizer = (
        torch.optim.Adam(gpmodule.parameters(), lr=0.01)
        if optimizer is None
        else optimizer
    )
    # TODO: add support for JIT loss
    loss_fn = TraceMeanField_ELBO().differentiable_loss if loss_fn is None else loss_fn

    def closure():
        optimizer.zero_grad()
        loss = loss_fn(gpmodule.model, gpmodule.guide)
        torch_backward(loss, retain_graph)
        return loss

    losses = []
    for i in range(num_steps):
        loss = optimizer.step(closure)
        losses.append(torch_item(loss))
    return losses



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import math
import re
import textwrap

import torch

from pyro.distributions.torch_distribution import TorchDistributionMixin
from pyro.distributions.util import broadcast_shape, sum_rightmost
from pyro.ops.special import log_binomial

from .. import settings
from . import constraints


def _clamp_by_zero(x):
    # works like clamp(x, min=0) but has grad at 0 is 0.5
    return (x.clamp(min=0) + x - x.clamp(max=0)) / 2


class Beta(torch.distributions.Beta, TorchDistributionMixin):
    def conjugate_update(self, other):
        """
        EXPERIMENTAL.
        """
        assert isinstance(other, Beta)
        concentration1 = self.concentration1 + other.concentration1 - 1
        concentration0 = self.concentration0 + other.concentration0 - 1
        updated = Beta(concentration1, concentration0)

        def _log_normalizer(d):
            x = d.concentration1
            y = d.concentration0
            return (x + y).lgamma() - x.lgamma() - y.lgamma()

        log_normalizer = (
            _log_normalizer(self) + _log_normalizer(other) - _log_normalizer(updated)
        )
        return updated, log_normalizer


class Binomial(torch.distributions.Binomial, TorchDistributionMixin):
    # EXPERIMENTAL threshold on total_count above which sampling will use a
    # clamped Poisson approximation for Binomial samples. This is useful for
    # sampling very large populations.
    approx_sample_thresh = math.inf

    # EXPERIMENTAL If set to a positive value, the .log_prob() method will use
    # a shifted Sterling's approximation to the Beta function, reducing
    # computational cost from 3 lgamma() evaluations to 4 log() evaluations
    # plus arithmetic. Recommended values are between 0.1 and 0.01.
    approx_log_prob_tol = 0.0

    def sample(self, sample_shape=torch.Size()):
        if self.approx_sample_thresh < math.inf:
            exact = self.total_count <= self.approx_sample_thresh
            if not exact.all():
                # Approximate large counts with a moment-matched clamped Poisson.
                with torch.no_grad():
                    shape = self._extended_shape(sample_shape)
                    p = self.probs
                    q = 1 - self.probs
                    mean = torch.min(p, q) * self.total_count
                    variance = p * q * self.total_count
                    shift = (mean - variance).round()
                    result = torch.poisson(variance.expand(shape))
                    result = torch.min(result + shift, self.total_count)
                    sample = torch.where(p < q, result, self.total_count - result)
                # Draw exact samples for remaining items.
                if exact.any():
                    total_count = torch.where(
                        exact, self.total_count, torch.zeros_like(self.total_count)
                    )
                    exact_sample = torch.distributions.Binomial(
                        total_count, self.probs, validate_args=False
                    ).sample(sample_shape)
                    sample = torch.where(exact, exact_sample, sample)
                return sample
        return super().sample(sample_shape)

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)

        n = self.total_count
        k = value
        # k * log(p) + (n - k) * log(1 - p) = k * (log(p) - log(1 - p)) + n * log(1 - p)
        #     (case logit < 0)              = k * logit - n * log1p(e^logit)
        #     (case logit > 0)              = k * logit - n * (log(p) - log(1 - p)) + n * log(p)
        #                                   = k * logit - n * logit - n * log1p(e^-logit)
        #     (merge two cases)             = k * logit - n * max(logit, 0) - n * log1p(e^-|logit|)
        normalize_term = n * (
            _clamp_by_zero(self.logits) + self.logits.abs().neg().exp().log1p()
        )
        return (
            k * self.logits
            - normalize_term
            + log_binomial(n, k, tol=self.approx_log_prob_tol)
        )


@settings.register(
    "binomial_approx_sample_thresh", __name__, "Binomial.approx_sample_thresh"
)
def _validate_thresh(thresh):
    assert isinstance(thresh, float)
    assert 0 < thresh


@settings.register(
    "binomial_approx_log_prob_tol", __name__, "Binomial.approx_log_prob_tol"
)
def _validate_tol(tol):
    assert isinstance(tol, float)
    assert 0 <= tol


# This overloads .log_prob() and .enumerate_support() to speed up evaluating
# log_prob on the support of this variable: we can completely avoid tensor ops
# and merely reshape the self.logits tensor. This is especially important for
# Pyro models that use enumeration.
class Categorical(torch.distributions.Categorical, TorchDistributionMixin):
    arg_constraints = {"probs": constraints.simplex, "logits": constraints.real_vector}

    def log_prob(self, value):
        if getattr(value, "_pyro_categorical_support", None) == id(self):
            # Assume value is a reshaped torch.arange(event_shape[0]).
            # In this case we can call .reshape() rather than torch.gather().
            if not torch._C._get_tracing_state():
                if self._validate_args:
                    self._validate_sample(value)
                assert value.size(0) == self.logits.size(-1)
            logits = self.logits
            if logits.dim() <= value.dim():
                logits = logits.reshape(
                    (1,) * (1 + value.dim() - logits.dim()) + logits.shape
                )
            if not torch._C._get_tracing_state():
                assert logits.size(-1 - value.dim()) == 1
            return logits.transpose(-1 - value.dim(), -1).squeeze(-1)
        return super().log_prob(value)

    def enumerate_support(self, expand=True):
        result = super().enumerate_support(expand=expand)
        if not expand:
            result._pyro_categorical_support = id(self)
        return result


class Dirichlet(torch.distributions.Dirichlet, TorchDistributionMixin):
    @staticmethod
    def infer_shapes(concentration):
        batch_shape = concentration[:-1]
        event_shape = concentration[-1:]
        return batch_shape, event_shape

    def conjugate_update(self, other):
        """
        EXPERIMENTAL.
        """
        assert isinstance(other, Dirichlet)
        concentration = self.concentration + other.concentration - 1
        updated = Dirichlet(concentration)

        def _log_normalizer(d):
            c = d.concentration
            return c.sum(-1).lgamma() - c.lgamma().sum(-1)

        log_normalizer = (
            _log_normalizer(self) + _log_normalizer(other) - _log_normalizer(updated)
        )
        return updated, log_normalizer


class Gamma(torch.distributions.Gamma, TorchDistributionMixin):
    def conjugate_update(self, other):
        """
        EXPERIMENTAL.
        """
        assert isinstance(other, Gamma)
        concentration = self.concentration + other.concentration - 1
        rate = self.rate + other.rate
        updated = Gamma(concentration, rate)

        def _log_normalizer(d):
            c = d.concentration
            return d.rate.log() * c - c.lgamma()

        log_normalizer = (
            _log_normalizer(self) + _log_normalizer(other) - _log_normalizer(updated)
        )
        return updated, log_normalizer


class Geometric(torch.distributions.Geometric, TorchDistributionMixin):
    # TODO: move upstream
    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)
        return (-value - 1) * torch.nn.functional.softplus(self.logits) + self.logits


class LogNormal(torch.distributions.LogNormal, TorchDistributionMixin):
    def __init__(self, loc, scale, validate_args=None):
        base_dist = Normal(loc, scale)
        # This differs from torch.distributions.LogNormal only in that base_dist is
        # a pyro.distributions.Normal rather than a torch.distributions.Normal.
        super(torch.distributions.LogNormal, self).__init__(
            base_dist,
            torch.distributions.transforms.ExpTransform(),
            validate_args=validate_args,
        )

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(LogNormal, _instance)
        return super(torch.distributions.LogNormal, self).expand(
            batch_shape, _instance=new
        )


class LowRankMultivariateNormal(
    torch.distributions.LowRankMultivariateNormal, TorchDistributionMixin
):
    @staticmethod
    def infer_shapes(loc, cov_factor, cov_diag):
        event_shape = loc[-1:]
        batch_shape = broadcast_shape(loc[:-1], cov_factor[:-2], cov_diag[:-1])
        return batch_shape, event_shape


class MultivariateNormal(
    torch.distributions.MultivariateNormal, TorchDistributionMixin
):
    @staticmethod
    def infer_shapes(
        loc, covariance_matrix=None, precision_matrix=None, scale_tril=None
    ):
        batch_shape, event_shape = loc[:-1], loc[-1:]
        for matrix in [covariance_matrix, precision_matrix, scale_tril]:
            if matrix is not None:
                batch_shape = broadcast_shape(batch_shape, matrix[:-2])
        return batch_shape, event_shape


class Multinomial(torch.distributions.Multinomial, TorchDistributionMixin):
    def infer_shapes(total_count=None, probs=None, logits=None):
        tensor = probs if logits is None else logits
        batch_shape, event_shape = tensor[:-1], tensor[-1:]
        if isinstance(total_count, tuple):
            batch_shape = broadcast_shape(batch_shape, total_count)
        return batch_shape, event_shape


class Normal(torch.distributions.Normal, TorchDistributionMixin):
    pass


class OneHotCategorical(torch.distributions.OneHotCategorical, TorchDistributionMixin):
    @staticmethod
    def infer_shapes(probs=None, logits=None):
        tensor = probs if logits is None else logits
        event_shape = tensor[-1:]
        batch_shape = tensor[:-1]
        return batch_shape, event_shape


class Poisson(torch.distributions.Poisson, TorchDistributionMixin):
    def __init__(self, rate, *, is_sparse=False, validate_args=None):
        self.is_sparse = is_sparse
        super().__init__(rate, validate_args=validate_args)

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(Poisson, _instance)
        new = super().expand(batch_shape, _instance=new)
        new.is_sparse = self.is_sparse
        return new

    def log_prob(self, value):
        if not self.is_sparse:
            return super().log_prob(value)
        if self._validate_args:
            self._validate_sample(value)
        rate, value, nonzero = torch.broadcast_tensors(self.rate, value, value > 0)
        sparse_rate = rate[nonzero]
        sparse_value = value[nonzero]
        return (
            torch.zeros_like(rate).masked_scatter(
                nonzero,
                (sparse_rate.log() * sparse_value) - (sparse_value + 1).lgamma(),
            )
            - rate
        )


class Independent(torch.distributions.Independent, TorchDistributionMixin):
    @staticmethod
    def infer_shapes(**kwargs):
        raise NotImplementedError

    @property
    def _validate_args(self):
        return self.base_dist._validate_args

    @_validate_args.setter
    def _validate_args(self, value):
        self.base_dist._validate_args = value

    def conjugate_update(self, other):
        """
        EXPERIMENTAL.
        """
        n = self.reintepreted_batch_ndims
        updated, log_normalizer = self.base_dist.conjugate_update(other.to_event(-n))
        updated = updated.to_event(n)
        log_normalizer = sum_rightmost(log_normalizer, n)
        return updated, log_normalizer


class Uniform(torch.distributions.Uniform, TorchDistributionMixin):
    def __init__(self, low, high, validate_args=None):
        self._unbroadcasted_low = low
        self._unbroadcasted_high = high
        super().__init__(low, high, validate_args=validate_args)

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(Uniform, _instance)
        new = super().expand(batch_shape, _instance=new)
        new._unbroadcasted_low = self._unbroadcasted_low
        new._unbroadcasted_high = self._unbroadcasted_high
        return new

    @constraints.dependent_property(is_discrete=False, event_dim=0)
    def support(self):
        return constraints.interval(self._unbroadcasted_low, self._unbroadcasted_high)


def _cat_docstrings(*docstrings):
    result = "\n".join(textwrap.dedent(s.lstrip("\n")) for s in docstrings)
    result = re.sub("\n\n+", "\n\n", result)
    # Drop torch-specific lines.
    result = "".join(
        line for line in result.splitlines(keepends=True) if "xdoctest" not in line
    )
    return result


# Programmatically load all distributions from PyTorch.
__all__ = []
for _name, _Dist in torch.distributions.__dict__.items():
    if not isinstance(_Dist, type):
        continue
    if not issubclass(_Dist, torch.distributions.Distribution):
        continue
    if _Dist is torch.distributions.Distribution:
        continue

    try:
        _PyroDist = locals()[_name]
    except KeyError:
        _PyroDist = type(_name, (_Dist, TorchDistributionMixin), {})
        _PyroDist.__module__ = __name__
        locals()[_name] = _PyroDist

    _PyroDist.__doc__ = """
    Wraps :class:`{}.{}` with
    :class:`~pyro.distributions.torch_distribution.TorchDistributionMixin`.

    """.format(
        _Dist.__module__, _Dist.__name__
    )
    _PyroDist.__doc__ = _cat_docstrings(_PyroDist.__doc__, _Dist.__doc__)
    __all__.append(_name)


# Create sphinx documentation.
__doc__ = "\n\n".join(
    [
        """
    {0}
    ----------------------------------------------------------------
    .. autoclass:: pyro.distributions.{0}
    """.format(
            _name
        )
        for _name in sorted(__all__)
        # Work around sphinx autodoc error in case two InverseGamma's are defined:
        # "duplicate object description of pyro.distributions.InverseGamma"
        if _name != "InverseGamma"
    ]
)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import warnings
from typing import TYPE_CHECKING, Callable, Dict, Set, Union

from typing_extensions import Self

from pyro import params
from pyro.distributions.distribution import Distribution
from pyro.poutine.messenger import Messenger
from pyro.poutine.util import is_validation_enabled

if TYPE_CHECKING:
    from pyro.poutine.runtime import Message


class LiftMessenger(Messenger):
    """
    Given a stochastic function with param calls and a prior distribution,
    create a stochastic function where all param calls are replaced by sampling from prior.
    Prior should be a callable or a dict of names to callables.

    Consider the following Pyro program:

        >>> def model(x):
        ...     s = pyro.param("s", torch.tensor(0.5))
        ...     z = pyro.sample("z", dist.Normal(x, s))
        ...     return z ** 2
        >>> lifted_model = pyro.poutine.lift(model, prior={"s": dist.Exponential(0.3)})

    ``lift`` makes ``param`` statements behave like ``sample`` statements
    using the distributions in ``prior``.  In this example, site `s` will now behave
    as if it was replaced with ``s = pyro.sample("s", dist.Exponential(0.3))``:

        >>> tr = pyro.poutine.trace(lifted_model).get_trace(0.0)
        >>> tr.nodes["s"]["type"] == "sample"
        True
        >>> tr2 = pyro.poutine.trace(lifted_model).get_trace(0.0)
        >>> bool((tr2.nodes["s"]["value"] == tr.nodes["s"]["value"]).all())
        False

    :param fn: function whose parameters will be lifted to random values
    :param prior: prior function in the form of a Distribution or a dict of stochastic fns
    :returns: ``fn`` decorated with a :class:`~pyro.poutine.lift_messenger.LiftMessenger`
    """

    def __init__(
        self,
        prior: Union[Callable, Distribution, Dict[str, Union[Distribution, Callable]]],
    ) -> None:
        """
        :param prior: prior used to lift parameters. Prior can be of type
                      dict, pyro.distributions, or a python stochastic fn

        Constructor
        """
        super().__init__()
        self.prior = prior
        self._samples_cache: Dict[str, "Message"] = {}

    def __enter__(self) -> Self:
        self._samples_cache = {}
        if is_validation_enabled() and isinstance(self.prior, dict):
            self._param_hits: Set[str] = set()
            self._param_misses: Set[str] = set()
        return super().__enter__()

    def __exit__(self, *args, **kwargs) -> None:
        self._samples_cache = {}
        if is_validation_enabled() and isinstance(self.prior, dict):
            extra = set(self.prior) - self._param_hits
            if extra:
                warnings.warn(
                    "pyro.module prior did not find params ['{}']. "
                    "Did you instead mean one of ['{}']?".format(
                        "', '".join(extra), "', '".join(self._param_misses)
                    )
                )
        return super().__exit__(*args, **kwargs)

    def _pyro_sample(self, msg: "Message") -> None:
        return None

    def _pyro_param(self, msg: "Message") -> None:
        """
        Overrides the `pyro.param` call with samples sampled from the
        distribution specified in the prior. The prior can be a
        pyro.distributions object or a dict of distributions keyed
        on the param names. If the param name does not match the
        name the keys in the prior, that param name is unchanged.
        """
        assert msg["name"] is not None
        name = msg["name"]
        param_name = params.user_param_name(name)
        if isinstance(self.prior, dict):
            # prior is a dict of distributions
            if param_name in self.prior.keys():
                msg["fn"] = self.prior[param_name]
                msg["args"] = msg["args"][1:]
                if isinstance(msg["fn"], Distribution):
                    msg["args"] = ()
                    msg["kwargs"] = {}
                    msg["infer"] = {}
                if is_validation_enabled():
                    self._param_hits.add(param_name)
            else:
                if is_validation_enabled():
                    self._param_misses.add(param_name)
                return None
        elif isinstance(self.prior, Distribution):
            # prior is a distribution
            msg["fn"] = self.prior
            msg["args"] = ()
            msg["kwargs"] = {}
            msg["infer"] = {}
        elif callable(self.prior):
            # prior is a stochastic fn. block sample
            msg["stop"] = True
            msg["fn"] = self.prior
            msg["args"] = msg["args"][1:]
        else:
            raise TypeError("unreachable")
        msg["type"] = "sample"
        if name in self._samples_cache:
            # Multiple pyro.param statements with the same
            # name. Block the site and fix the value.
            msg["value"] = self._samples_cache[name]["value"]
            msg["is_observed"] = True
            msg["stop"] = True
        else:
            self._samples_cache[name] = msg
            msg["is_observed"] = False
        return self._pyro_sample(msg)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0


def clamp_preserve_gradients(x, min, max):
    # This helper function clamps gradients but still passes through the gradient in clamped regions
    return x + (x.clamp(min, max) - x).detach()



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import math

import torch

from pyro.contrib.oed.glmm import analytic_posterior_cov
from pyro.contrib.util import get_indices
from pyro.infer.autoguide.utils import mean_field_entropy


def linear_model_ground_truth(
    model, design, observation_labels, target_labels, eig=True
):
    if isinstance(target_labels, str):
        target_labels = [target_labels]

    w_sd = torch.cat(list(model.w_sds.values()), dim=-1)
    prior_cov = torch.diag(w_sd**2)
    design_shape = design.shape
    posterior_covs = [
        analytic_posterior_cov(prior_cov, x, model.obs_sd)
        for x in torch.unbind(design.reshape(-1, design_shape[-2], design_shape[-1]))
    ]
    target_indices = get_indices(target_labels, tensors=model.w_sds)
    target_posterior_covs = [
        S[target_indices, :][:, target_indices] for S in posterior_covs
    ]
    output = torch.tensor(
        [0.5 * torch.logdet(2 * math.pi * math.e * C) for C in target_posterior_covs]
    )
    if eig:
        prior_entropy = mean_field_entropy(model, [design], whitelist=target_labels)
        output = prior_entropy - output

    return output.reshape(design.shape[:-2])



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

from collections import defaultdict
from collections.abc import Iterable
from functools import singledispatch

import pyro
from pyro.poutine.handlers import _make_handler
from pyro.poutine.reentrant_messenger import ReentrantMessenger
from pyro.poutine.runtime import effectful


@effectful(type="genname")
def genname(name="name"):
    return name


class NameScope:
    def __init__(self, name=None):
        self.name = name
        self.counter = 0
        self._namespace = defaultdict(int)

    def __str__(self):
        if self.counter:
            return f"{self.name}__{self.counter}"
        return str(self.name)

    def allocate(self, name):
        counter = self._namespace[name]
        self._namespace[name] += 1
        return counter


class ScopeStack:
    """
    Single global state to keep track of scope stacks.
    """

    def __init__(self):
        self._stack = []

    def __str__(self):
        return "/".join(str(scope) for scope in self._stack)

    @property
    def global_scope(self):
        return NameScope()  # don't keep a counter for a global scope

    @property
    def current_scope(self):
        if len(self._stack):
            return self._stack[-1]
        return self.global_scope

    def push_scope(self, scope):
        scope.counter = self.current_scope.allocate(scope.name)
        self._stack.append(scope)

    def pop_scope(self):
        return self._stack.pop(-1)

    def fresh_name(self, name):
        counter = self.current_scope.allocate(name)
        if counter:
            return name + str(counter)
        return name


class AutonameMessenger(ReentrantMessenger):
    """
    Assign unique names to random variables.

    1. For a new varialbe use its declared name if given, otherwise use the distribution name::

        sample("x", dist.Bernoulli ... )  # -> x
        sample(dist.Bernoulli ... )  # -> Bernoulli

    2. For repeated variables names append the counter as a suffix::

        sample(dist.Bernoulli ... )  # -> Bernoulli
        sample(dist.Bernoulli ... )  # -> Bernoulli1
        sample(dist.Bernoulli ... )  # -> Bernoulli2

    3. Functions and iterators can be used as a name scope::

        @autoname
        def f1():
            sample(dist.Bernoulli ... )

        @autoname
        def f2():
            f1()  # -> f2/f1/Bernoulli
            f1()  # -> f2/f1__1/Bernoulli
            sample(dist.Bernoulli ... )  # -> f2/Bernoulli

        @autoname(name="model")
        def f3():
            for i in autoname(range(3), name="time"):
                # model/time/Bernoulli .. model/time__1/Bernoulli .. model/time__2/Bernoulli
                sample(dist.Bernoulli ... )
                # model/time/f1/Bernoulli .. model/time__1/f1/Bernoulli .. model/time__2/f1/Bernoulli
                f1()

    4. Or scopes can be added using the with statement::

        def f4():
            with autoname(name="prefix"):
                f1()  # -> prefix/f1/Bernoulli
                f1()  # -> prefix/f1__1/Bernoulli
                sample(dist.Bernoulli ... )  # -> prefix/Bernoulli
    """

    def __init__(self, name=None):
        self.name = name
        super().__init__()

    def __call__(self, fn_or_iter):
        if isinstance(fn_or_iter, Iterable):
            if self.name is None:
                self.name = fn_or_iter.name  # name of a sequential pyro.plate
            self._iter = fn_or_iter
            return self
        if callable(fn_or_iter):
            if self.name is None:
                self.name = fn_or_iter.__name__
            return super().__call__(fn_or_iter)
        raise ValueError(f"{fn_or_iter} has to be an iterable or a callable.")

    def __enter__(self):
        scope = NameScope(self.name)
        _SCOPE_STACK.push_scope(scope)
        return super().__enter__()

    def __exit__(self, *args):
        _SCOPE_STACK.pop_scope()
        return super().__exit__(*args)

    def __iter__(self):
        for i in self._iter:
            scope = NameScope(self.name)
            _SCOPE_STACK.push_scope(scope)
            yield i
            scope = _SCOPE_STACK.pop_scope()

    @staticmethod  # only depends on the global _SCOPE_STACK state, not self
    def _pyro_genname(msg):
        raw_name = msg["fn"](*msg["args"])
        fresh_name = _SCOPE_STACK.fresh_name(raw_name)

        msg["value"] = str(_SCOPE_STACK) + "/" + fresh_name
        msg["stop"] = True


@_make_handler(AutonameMessenger, __name__)
def autoname(fn=None, name=None): ...


@singledispatch
def sample(*args):
    raise NotImplementedError


@sample.register(str)
def _sample_name(name, fn, *args, **kwargs):  # the current syntax of pyro.sample
    name = genname(name)
    return pyro.sample(name, fn, *args, **kwargs)


@sample.register(pyro.distributions.Distribution)
def _sample_dist(fn, *args, **kwargs):
    name = kwargs.pop("name", None)
    name = genname(type(fn).__name__ if name is None else name)
    return pyro.sample(name, fn, *args, **kwargs)


_SCOPE_STACK = ScopeStack()



# Copyright Contributors to the Pyro project.
# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

from .conjugate import ConjugateReparam
from .discrete_cosine import DiscreteCosineReparam
from .haar import HaarReparam
from .hmm import LinearHMMReparam
from .loc_scale import LocScaleReparam
from .neutra import NeuTraReparam
from .projected_normal import ProjectedNormalReparam
from .softmax import GumbelSoftmaxReparam
from .split import SplitReparam
from .stable import LatentStableReparam, StableReparam, SymmetricStableReparam
from .strategies import AutoReparam, MinimalReparam
from .structured import StructuredReparam
from .studentt import StudentTReparam
from .transform import TransformReparam
from .unit_jacobian import UnitJacobianReparam

__all__ = [
    "AutoReparam",
    "ConjugateReparam",
    "DiscreteCosineReparam",
    "GumbelSoftmaxReparam",
    "HaarReparam",
    "LatentStableReparam",
    "LinearHMMReparam",
    "LocScaleReparam",
    "MinimalReparam",
    "NeuTraReparam",
    "ProjectedNormalReparam",
    "SplitReparam",
    "StableReparam",
    "StructuredReparam",
    "StudentTReparam",
    "SymmetricStableReparam",
    "TransformReparam",
    "UnitJacobianReparam",
]



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

"""
These reparametrization strategies are registered with
:func:`~pyro.poutine.reparam_messenger.register_reparam_strategy` and are
accessed by name via ``poutine.reparam(config=name_of_strategy)`` .
See :func:`~pyro.poutine.handlers.reparam` for usage.
"""

from abc import ABC, abstractmethod
from typing import Callable, Dict, Optional, Union

import torch
from torch.distributions import constraints

import pyro.distributions as dist
import pyro.poutine as poutine

from .loc_scale import LocScaleReparam
from .projected_normal import ProjectedNormalReparam
from .reparam import Reparam
from .softmax import GumbelSoftmaxReparam
from .stable import LatentStableReparam, StableReparam, SymmetricStableReparam
from .transform import TransformReparam


class Strategy(ABC):
    """
    Abstract base class for reparametrizer configuration strategies.

    Derived classes must implement the :meth:`configure` method.

    :ivar dict config: A dictionary configuration. This will be populated the
        first time the model is run. Thereafter it can be used as an argument
        to ``poutine.reparam(config=___)``.
    """

    # TODO(https://github.com/pyro-ppl/pyro/issues/2831) As part of refactoring
    # Reparam objects to be nn.Modules:
    # - make Strategy inherit from torch.nn.Module
    # - make self.config a torch.nn.ModuleDict

    def __init__(self):
        # TODO(#2831) Make this a torch.nn.ModuleDict.
        self.config: Dict[str, Optional[Reparam]] = {}
        super().__init__()

    @abstractmethod
    def configure(self, msg: dict) -> Optional[Reparam]:
        """
        Inputs a sample site and returns either None or a
        :class:`~pyro.infer.reparam.reparam.Reparam` instance.

        This will be called only on the first model execution; subsequent
        executions will use the reparametrizer stored in ``self.config``.

        :param dict msg: A sample site to possibly reparametrize.
        :returns: An optional reparametrizer instance.
        """
        raise NotImplementedError

    def __call__(self, msg_or_fn: Union[dict, Callable]):
        """
        Strategies can be used as decorators to reparametrize a model.

        :param msg_or_fn: Public use: a model to be decorated. (Internal use: a
            site to be configured for reparametrization).
        """
        if isinstance(msg_or_fn, dict):  # Internal use during configuration.
            msg = msg_or_fn
            name = msg["name"]
            if name in self.config:
                return self.config[name]
            result = self.configure(msg)
            self.config[name] = result
            return result
        else:  # Public use as a decorator or handler.
            fn = msg_or_fn
            return poutine.reparam(fn, self)


class MinimalReparam(Strategy):
    """
    Minimal reparametrization strategy that reparametrizes only those sites
    that would otherwise lead to error, e.g.
    :class:`~pyro.distributions.Stable` and
    :class:`~pyro.distributions.ProjectedNormal` random variables.

    Example::

        @MinimalReparam()
        def model(...):
            ...

    which is equivalent to::

        @poutine.reparam(config=MinimalReparam())
        def model(...):
            ...
    """

    def configure(self, msg: dict) -> Optional[Reparam]:
        return _minimal_reparam(msg["fn"], msg["is_observed"])


def _minimal_reparam(fn, is_observed):
    # Unwrap Independent, Masked, Transformed etc.
    while isinstance(getattr(fn, "base_dist", None), dist.Distribution):
        if isinstance(fn, torch.distributions.TransformedDistribution):
            if _minimal_reparam(fn.base_dist, is_observed) is None:
                return None  # No need to reparametrize.
            else:
                return TransformReparam()  # Then reparametrize new sites.
        fn = fn.base_dist

    if isinstance(fn, dist.Stable):
        if not is_observed:
            return LatentStableReparam()
        elif fn.skew.requires_grad or fn.skew.any():
            return StableReparam()
        else:
            return SymmetricStableReparam()

    if isinstance(fn, dist.ProjectedNormal):
        return ProjectedNormalReparam()

    # TODO apply CircularReparam for VonMises


class AutoReparam(Strategy):
    """
    Applies a recommended set of reparametrizers. These currently include:
    :class:`MinimalReparam`,
    :class:`~pyro.infer.reparam.transform.TransformReparam`, a fully-learnable
    :class:`~pyro.infer.reparam.loc_scale.LocScaleReparam`, and
    :class:`~pyro.infer.reparam.softmax.GumbelSoftmaxReparam`.

    Example::

        @AutoReparam()
        def model(...):
            ...

    which is equivalent to::

        @poutine.reparam(config=AutoReparam())
        def model(...):
            ...

    .. warning:: This strategy may change behavior across Pyro releases.
        To inspect or save a given behavior, extract the ``.config`` dict after
        running the model at least once.

    :param centered: Optional centering parameter for
        :class:`~pyro.infer.reparam.loc_scale.LocScaleReparam` reparametrizers.
        If None (default), centering will be learned. If a float in
        ``[0.0,1.0]``, then a fixed centering. To completely decenter (e.g. in
        MCMC), set to 0.0.
    """

    def __init__(self, *, centered: Optional[float] = None):
        assert centered is None or isinstance(centered, float)
        super().__init__()
        self.centered = centered

    def configure(self, msg: dict) -> Optional[Reparam]:
        # Focus on tricks for latent sites.
        fn = msg["fn"]
        if not msg["is_observed"]:
            # Unwrap Independent, Masked, Transformed etc.
            while isinstance(getattr(fn, "base_dist", None), dist.Distribution):
                if isinstance(fn, torch.distributions.TransformedDistribution):
                    return TransformReparam()  # Then reparametrize new sites.
                fn = fn.base_dist

            # Try to apply a GumbelSoftmaxReparam.
            if isinstance(fn, torch.distributions.RelaxedOneHotCategorical):
                return GumbelSoftmaxReparam()

            # Apply a learnable LocScaleReparam.
            result = _loc_scale_reparam(msg["name"], fn, self.centered)
            if result is not None:
                return result

        # Apply minimal reparametrizers.
        return _minimal_reparam(fn, msg["is_observed"])


def _loc_scale_reparam(name, fn, centered):
    if "_decentered" in name:
        return  # Avoid infinite recursion.

    # Check for location-scale families.
    params = set(fn.arg_constraints)
    if not {"loc", "scale"}.issubset(params):
        return

    # Check for unconstrained support.
    if not _is_unconstrained(fn.support):
        return

    # TODO reparametrize only if parameters are variable. We might guess
    # based on whether parameters are differentiable, .requires_grad. See
    # https://github.com/pyro-ppl/pyro/pull/2824

    # Create an elementwise-learnable reparametrizer.
    shape_params = sorted(params - {"loc", "scale"})
    return LocScaleReparam(centered=centered, shape_params=shape_params)


def _is_unconstrained(constraint):
    # Unwrap constraints.independent.
    while hasattr(constraint, "base_constraint"):
        constraint = constraint.base_constraint
    return constraint == constraints.real



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import torch

from pyro.distributions import constraints
from pyro.distributions.torch import Categorical


class OrderedLogistic(Categorical):
    """
    Alternative parametrization of the distribution over a categorical variable.

    Instead of the typical parametrization of a categorical variable in terms
    of the probability mass of the individual categories ``p``, this provides an
    alternative that is useful in specifying ordered categorical models. This
    accepts a vector of ``cutpoints`` which are an ordered vector of real numbers
    denoting baseline cumulative log-odds of the individual categories, and a
    model vector ``predictor`` which modifies the baselines for each sample
    individually.

    These cumulative log-odds are then transformed into a discrete cumulative
    probability distribution, that is finally differenced to return the probability
    mass matrix ``p`` that specifies the categorical distribution.

    :param Tensor predictor: A tensor of predictor variables of arbitrary
        shape. The output shape of non-batched samples from this distribution will
        be the same shape as ``predictor``.
    :param Tensor cutpoints: A tensor of cutpoints that are used to determine the
        cumulative probability of each entry in ``predictor`` belonging to a
        given category. The first `cutpoints.ndim-1` dimensions must be
        broadcastable to ``predictor``, and the -1 dimension is monotonically
        increasing.
    """

    arg_constraints = {
        "predictor": constraints.real,
        "cutpoints": constraints.ordered_vector,
    }

    def __init__(self, predictor, cutpoints, validate_args=None):
        # calculate cumulative probability for each predictor
        q = torch.sigmoid(cutpoints - predictor.unsqueeze(-1))
        # expand parameters to match batch shape
        self.predictor = predictor.expand(q.shape[:-1])
        self.cutpoints = cutpoints.expand(q.shape)
        # turn cumulative probabilities into probability mass of categories
        p_shape = q.shape[:-1] + (q.shape[-1] + 1,)
        p = torch.zeros(p_shape, dtype=q.dtype, device=q.device)
        p[..., 0] = q[..., 0]
        p[..., 1:-1] = q[..., 1:] - q[..., :-1]
        p[..., -1] = 1 - q[..., -1]
        # pass probability mass to Categorical constructor
        super(OrderedLogistic, self).__init__(p, validate_args=validate_args)

    def expand(self, batch_shape, _instance=None):
        batch_shape = torch.Size(batch_shape)
        new = self._get_checked_instance(OrderedLogistic, _instance)
        new.predictor = self.predictor.expand(batch_shape)
        new.cutpoints = self.cutpoints.expand(batch_shape + (self.cutpoints.shape[-1],))
        return super(OrderedLogistic, self).expand(batch_shape, new)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import copy
import warnings
from collections import OrderedDict
from contextlib import ExitStack, contextmanager
from inspect import isclass
from typing import Callable, Iterator, Optional, Sequence, Union

import torch
from torch.distributions import constraints

import pyro.distributions as dist
import pyro.infer as infer
import pyro.poutine as poutine
from pyro.distributions.torch_distribution import TorchDistributionMixin
from pyro.params import param_with_module_name
from pyro.params.param_store import ParamStoreDict
from pyro.poutine.plate_messenger import PlateMessenger
from pyro.poutine.runtime import (
    _MODULE_NAMESPACE_DIVIDER,
    _PYRO_PARAM_STORE,
    InferDict,
    Message,
    am_i_wrapped,
    apply_stack,
    effectful,
)
from pyro.poutine.subsample_messenger import SubsampleMessenger
from pyro.util import deep_getattr, set_rng_seed  # noqa: F401


def get_param_store() -> ParamStoreDict:
    """
    Returns the global :class:`~pyro.params.param_store.ParamStoreDict`.
    """
    return _PYRO_PARAM_STORE


def clear_param_store() -> None:
    """
    Clears the global :class:`~pyro.params.param_store.ParamStoreDict`.

    This is especially useful if you're working in a REPL. We recommend calling
    this before each training loop (to avoid leaking parameters from past
    models), and before each unit test (to avoid leaking parameters across
    tests).
    """
    _PYRO_PARAM_STORE.clear()


_param = effectful(_PYRO_PARAM_STORE.get_param, type="param")


def param(
    name: str,
    init_tensor: Union[torch.Tensor, Callable[[], torch.Tensor], None] = None,
    constraint: constraints.Constraint = constraints.real,
    event_dim: Optional[int] = None,
) -> torch.Tensor:
    """
    Saves the variable as a parameter in the param store.
    To interact with the param store or write to disk,
    see `Parameters <parameters.html>`_.

    :param str name: name of parameter
    :param init_tensor: initial tensor or lazy callable that returns a tensor.
        For large tensors, it may be cheaper to write e.g.
        ``lambda: torch.randn(100000)``, which will only be evaluated on the
        initial statement.
    :type init_tensor: torch.Tensor or callable
    :param constraint: torch constraint, defaults to ``constraints.real``.
    :type constraint: torch.distributions.constraints.Constraint
    :param int event_dim: (optional) number of rightmost dimensions unrelated
        to batching. Dimension to the left of this will be considered batch
        dimensions; if the param statement is inside a subsampled plate, then
        corresponding batch dimensions of the parameter will be correspondingly
        subsampled. If unspecified, all dimensions will be considered event
        dims and no subsampling will be performed.
    :returns: A constrained parameter. The underlying unconstrained parameter
        is accessible via ``pyro.param(...).unconstrained()``, where
        ``.unconstrained`` is a weakref attribute.
    :rtype: torch.Tensor
    """
    # Note effectful(-) requires the double passing of name below.
    args = (name,) if init_tensor is None else (name, init_tensor)
    value = _param(*args, constraint=constraint, event_dim=event_dim, name=name)
    assert value is not None  # type narrowing guaranteed by _param
    return value


def _masked_observe(
    name: str,
    fn: TorchDistributionMixin,
    obs: Optional[torch.Tensor],
    obs_mask: torch.BoolTensor,
    *args,
    **kwargs,
) -> torch.Tensor:
    # Split into two auxiliary sample sites.
    with poutine.mask(mask=obs_mask):
        observed = sample(f"{name}_observed", fn, *args, **kwargs, obs=obs)
    with poutine.mask(mask=~obs_mask):  # type: ignore[call-overload]
        unobserved = sample(f"{name}_unobserved", fn, *args, **kwargs)

    # Interleave observed and unobserved events.
    shape = obs_mask.shape + (1,) * fn.event_dim
    batch_mask = obs_mask.reshape(shape)
    try:
        value = torch.where(batch_mask, observed, unobserved)
    except RuntimeError as e:
        if "must match the size of tensor" in str(e):
            shape = torch.broadcast_shapes(observed.shape, unobserved.shape)
            batch_shape = shape[: len(shape) - fn.event_dim]
            raise ValueError(
                f"Invalid obs_mask shape {tuple(obs_mask.shape)}; should be "
                f"broadcastable to batch_shape = {tuple(batch_shape)}"
            ) from e
        raise
    return deterministic(name, value, event_dim=fn.event_dim)


def sample(
    name: str,
    fn: TorchDistributionMixin,
    *args,
    obs: Optional[torch.Tensor] = None,
    obs_mask: Optional[torch.BoolTensor] = None,
    infer: Optional[InferDict] = None,
    **kwargs,
) -> torch.Tensor:
    """
    Calls the stochastic function ``fn`` with additional side-effects depending
    on ``name`` and the enclosing context (e.g. an inference algorithm).  See
    `Introduction to Pyro <http://pyro.ai/examples/intro_long.html>`_ for a discussion.

    :param name: name of sample
    :param fn: distribution class or function
    :param obs: observed datum (optional; should only be used in context of
        inference) optionally specified in kwargs
    :param ~torch.Tensor obs_mask: Optional boolean tensor mask of shape
        broadcastable with ``fn.batch_shape``. If provided, events with
        mask=True will be conditioned on ``obs`` and remaining events will be
        imputed by sampling. This introduces a latent sample site named ``name
        + "_unobserved"`` which should be used by guides.
    :type obs_mask: bool or ~torch.Tensor
    :param dict infer: Optional dictionary of inference parameters specified
        in kwargs. See inference documentation for details.
    :returns: sample
    """
    # Transform obs_mask into multiple sample statements.
    if obs_mask is not None:
        return _masked_observe(name, fn, obs, obs_mask, *args, **kwargs)

    # Check if stack is empty.
    # if stack empty, default behavior (defined here)
    infer = {} if infer is None else infer.copy()
    is_observed = infer.pop("is_observed", obs is not None)
    assert isinstance(is_observed, bool)
    if not am_i_wrapped():
        if obs is not None and not infer.get("_deterministic"):
            warnings.warn(
                "trying to observe a value outside of inference at " + name,
                RuntimeWarning,
            )
            return obs
        return fn(*args, **kwargs)
    # if stack not empty, apply everything in the stack?
    else:
        # initialize data structure to pass up/down the stack
        msg = Message(
            type="sample",
            name=name,
            fn=fn,
            is_observed=is_observed,
            args=args,
            kwargs=kwargs,
            value=obs,
            infer=infer,
            scale=1.0,
            mask=None,
            cond_indep_stack=(),
            done=False,
            stop=False,
            continuation=None,
        )
        # apply the stack and return its return value
        apply_stack(msg)
        assert msg["value"] is not None  # type narrowing guaranteed by apply_stack
        return msg["value"]


def factor(
    name: str, log_factor: torch.Tensor, *, has_rsample: Optional[bool] = None
) -> None:
    """
    Factor statement to add arbitrary log probability factor to a
    probabilisitic model.

    .. warning:: When using factor statements in guides, you'll need to specify
        whether the factor statement originated from fully reparametrized
        sampling (e.g. the Jacobian determinant of a transformation of a
        reparametrized variable) or from nonreparameterized sampling (e.g.
        discrete samples). For the fully reparametrized case, set
        ``has_rsample=True``; for the nonreparametrized case, set
        ``has_rsample=False``. This is needed only in guides, not in models.

    :param str name: Name of the trivial sample
    :param torch.Tensor log_factor: A possibly batched log probability factor.
    :param bool has_rsample: Whether the ``log_factor`` arose from a fully
        reparametrized distribution. Defaults to False when used in models, but
        must be specified for use in guides.
    """
    unit_dist = dist.Unit(log_factor, has_rsample=has_rsample)
    unit_value = unit_dist.sample()
    sample(name, unit_dist, obs=unit_value, infer={"is_auxiliary": True})


def deterministic(
    name: str, value: torch.Tensor, event_dim: Optional[int] = None
) -> torch.Tensor:
    """
    Deterministic statement to add a :class:`~pyro.distributions.Delta` site
    with name `name` and value `value` to the trace. This is useful when we
    want to record values which are completely determined by their parents.
    For example::

        x = pyro.sample("x", dist.Normal(0, 1))
        x2 = pyro.deterministic("x2", x ** 2)

    .. note:: The site does not affect the model density. This currently converts
        to a :func:`sample` statement, but may change in the future.

    :param str name: Name of the site.
    :param torch.Tensor value: Value of the site.
    :param int event_dim: Optional event dimension, defaults to `value.ndim`.
    """
    event_dim = value.ndim if event_dim is None else event_dim
    return sample(
        name,
        dist.Delta(value, event_dim=event_dim).mask(False),
        obs=value,
        infer={"_deterministic": True},
    )


@effectful(type="subsample")
def subsample(data: torch.Tensor, event_dim: int) -> torch.Tensor:
    """
    Subsampling statement to subsample data tensors based on enclosing
    :class:`plate` s.

    This is typically called on arguments to ``model()`` when subsampling is
    performed automatically by :class:`plate` s by passing either the
    ``subsample`` or ``subsample_size`` kwarg. For example the following are
    equivalent::

        # Version 1. using indexing
        def model(data):
            with pyro.plate("data", len(data), subsample_size=10, dim=-data.dim()) as ind:
                data = data[ind]
                # ...

        # Version 2. using pyro.subsample()
        def model(data):
            with pyro.plate("data", len(data), subsample_size=10, dim=-data.dim()):
                data = pyro.subsample(data, event_dim=0)
                # ...

    :param data: A tensor of batched data.
    :type data: ~torch.Tensor
    :param int event_dim: The event dimension of the data tensor. Dimensions to
        the left are considered batch dimensions.
    :returns: A subsampled version of ``data``
    :rtype: ~torch.Tensor
    """
    assert isinstance(event_dim, int) and event_dim >= 0
    return data  # May be intercepted by SubsampleMessenger.


class plate(PlateMessenger):
    """
    Construct for conditionally independent sequences of variables.

    ``plate`` can be used either sequentially as a generator or in parallel as
    a context manager (formerly ``irange`` and ``iarange``, respectively).

    Sequential :class:`plate` is similar to :py:func:`range` in that it generates
    a sequence of values.

    Vectorized :class:`plate` is similar to :func:`torch.arange` in that it
    yields an array of indices by which other tensors can be indexed.
    :class:`plate` differs from :func:`torch.arange` in that it also informs
    inference algorithms that the variables being indexed are conditionally
    independent.  To do this, :class:`plate` is a provided as context manager
    rather than a function, and users must guarantee that all computation
    within an :class:`plate` context is conditionally independent::

        with pyro.plate("name", size) as ind:
            # ...do conditionally independent stuff with ind...

    Additionally, :class:`plate` can take advantage of the conditional
    independence assumptions by subsampling the indices and informing inference
    algorithms to scale various computed values. This is typically used to
    subsample minibatches of data::

        with pyro.plate("data", len(data), subsample_size=100) as ind:
            batch = data[ind]
            assert len(batch) == 100

    By default ``subsample_size=False`` and this simply yields a
    ``torch.arange(0, size)``. If ``0 < subsample_size <= size`` this yields a
    single random batch of indices of size ``subsample_size`` and scales all
    log likelihood terms by ``size/batch_size``, within this context.

    .. warning::  This is only correct if all computation is conditionally
        independent within the context.

    :param str name: A unique name to help inference algorithms match
        :class:`plate` sites between models and guides.
    :param int size: Optional size of the collection being subsampled
        (like `stop` in builtin `range`).
    :param int subsample_size: Size of minibatches used in subsampling.
        Defaults to `size`.
    :param subsample: Optional custom subsample for user-defined subsampling
        schemes. If specified, then `subsample_size` will be set to
        `len(subsample)`.
    :type subsample: Anything supporting `len()`.
    :param int dim: An optional dimension to use for this independence index.
        If specified, ``dim`` should be negative, i.e. should index from the
        right. If not specified, ``dim`` is set to the rightmost dim that is
        left of all enclosing ``plate`` contexts.
    :param bool use_cuda: DEPRECATED, use the `device` arg instead.
        Optional bool specifying whether to use cuda tensors for `subsample`
        and `log_prob`. Defaults to ``torch.Tensor.is_cuda``.
    :param str device: Optional keyword specifying which device to place
        the results of `subsample` and `log_prob` on. By default, results
        are placed on the same device as the default tensor.
    :return: A reusabe context manager yielding a single 1-dimensional
        :class:`torch.Tensor` of indices.

    Examples:

        .. doctest::
           :hide:

           >>> loc, scale = torch.tensor(0.), torch.tensor(1.)
           >>> data = torch.randn(100)
           >>> z = dist.Bernoulli(0.5).sample((100,))

        >>> # This version declares sequential independence and subsamples data:
        >>> for i in pyro.plate('data', 100, subsample_size=10):
        ...     if z[i]:  # Control flow in this example prevents vectorization.
        ...         obs = pyro.sample(f'obs_{i}', dist.Normal(loc, scale),
        ...                           obs=data[i])

        >>> # This version declares vectorized independence:
        >>> with pyro.plate('data'):
        ...     obs = pyro.sample('obs', dist.Normal(loc, scale), obs=data)

        >>> # This version subsamples data in vectorized way:
        >>> with pyro.plate('data', 100, subsample_size=10) as ind:
        ...     obs = pyro.sample('obs', dist.Normal(loc, scale), obs=data[ind])

        >>> # This wraps a user-defined subsampling method for use in pyro:
        >>> ind = torch.randint(0, 100, (10,)).long() # custom subsample
        >>> with pyro.plate('data', 100, subsample=ind):
        ...     obs = pyro.sample('obs', dist.Normal(loc, scale), obs=data[ind])

        >>> # This reuses two different independence contexts.
        >>> x_axis = pyro.plate('outer', 320, dim=-1)
        >>> y_axis = pyro.plate('inner', 200, dim=-2)
        >>> with x_axis:
        ...     x_noise = pyro.sample("x_noise", dist.Normal(loc, scale))
        ...     assert x_noise.shape == (320,)
        >>> with y_axis:
        ...     y_noise = pyro.sample("y_noise", dist.Normal(loc, scale))
        ...     assert y_noise.shape == (200, 1)
        >>> with x_axis, y_axis:
        ...     xy_noise = pyro.sample("xy_noise", dist.Normal(loc, scale))
        ...     assert xy_noise.shape == (200, 320)

    See `SVI Part II <http://pyro.ai/examples/svi_part_ii.html>`_ for an
    extended discussion.
    """

    pass


class iarange(plate):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "pyro.iarange is deprecated; use pyro.plate instead", DeprecationWarning
        )
        super().__init__(*args, **kwargs)


class irange(SubsampleMessenger):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "pyro.irange is deprecated; use pyro.plate instead", DeprecationWarning
        )
        super().__init__(*args, **kwargs)


@contextmanager
def plate_stack(
    prefix: str, sizes: Sequence[int], rightmost_dim: int = -1
) -> Iterator[None]:
    """
    Create a contiguous stack of :class:`plate` s with dimensions::

        rightmost_dim - len(sizes), ..., rightmost_dim

    :param str prefix: Name prefix for plates.
    :param iterable sizes: An iterable of plate sizes.
    :param int rightmost_dim: The rightmost dim, counting from the right.
    """
    assert rightmost_dim < 0
    with ExitStack() as stack:
        for i, size in enumerate(reversed(sizes)):
            plate_i = plate(f"{prefix}_{i}", size, dim=rightmost_dim - i)
            stack.enter_context(plate_i)
        yield


def module(
    name: str, nn_module: torch.nn.Module, update_module_params: bool = False
) -> torch.nn.Module:
    """
    Registers all parameters of a :class:`torch.nn.Module` with Pyro's
    :mod:`~pyro.params.param_store`.  In conjunction with the
    :class:`~pyro.params.param_store.ParamStoreDict`
    :meth:`~pyro.params.param_store.ParamStoreDict.save` and
    :meth:`~pyro.params.param_store.ParamStoreDict.load` functionality, this
    allows the user to save and load modules.

    .. note:: Consider instead using :class:`~pyro.nn.module.PyroModule`, a
        newer alternative to ``pyro.module()`` that has better support for:
        jitting, serving in C++, and converting parameters to random variables.
        For details see the `Modules Tutorial
        <https://pyro.ai/examples/modules.html>`_ .

    :param name: name of module
    :type name: str
    :param nn_module: the module to be registered with Pyro
    :type nn_module: torch.nn.Module
    :param update_module_params: determines whether Parameters
                                 in the PyTorch module get overridden with the values found in the
                                 ParamStore (if any). Defaults to `False`
    :type load_from_param_store: bool
    :returns: torch.nn.Module
    """
    assert hasattr(nn_module, "parameters"), "module has no parameters"
    assert _MODULE_NAMESPACE_DIVIDER not in name, (
        "improper module name, since contains %s" % _MODULE_NAMESPACE_DIVIDER
    )

    if isclass(nn_module):
        raise NotImplementedError(
            "pyro.module does not support class constructors for "
            + "the argument nn_module"
        )

    target_state_dict = OrderedDict()

    for param_name, param_value in nn_module.named_parameters():
        if param_value.requires_grad:
            # register the parameter in the module with pyro
            # this only does something substantive if the parameter hasn't been seen before
            full_param_name = param_with_module_name(name, param_name)
            returned_param = param(full_param_name, param_value)

            if param_value._cdata != returned_param._cdata:
                target_state_dict[param_name] = returned_param
        elif nn_module.training:
            warnings.warn(
                f"{param_name} was not registered in the param store "
                "because requires_grad=False. You can silence this "
                "warning by calling my_module.train(False)"
            )

    if target_state_dict and update_module_params:
        # WARNING: this is very dangerous. better method?
        for _name, _param in nn_module.named_parameters():
            is_param = False
            name_arr = _name.rsplit(".", 1)
            if len(name_arr) > 1:
                mod_name, param_name = name_arr[0], name_arr[1]
            else:
                is_param = True
                mod_name = _name
            if _name in target_state_dict.keys():
                if not is_param:
                    deep_getattr(nn_module, mod_name)._parameters[param_name] = (
                        target_state_dict[_name]
                    )
                else:
                    nn_module._parameters[mod_name] = target_state_dict[_name]  # type: ignore[assignment]

    return nn_module


def random_module(name, nn_module, prior, *args, **kwargs):
    r"""
    .. warning::
        The `random_module` primitive is deprecated, and will be removed
        in a future release. Use :class:`~pyro.nn.module.PyroModule` instead to
        to create Bayesian modules from :class:`torch.nn.Module` instances.
        See the `Bayesian Regression tutorial <http://pyro.ai/examples/bayesian_regression.html>`_
        for an example.

    DEPRECATED Places a prior over the parameters of the module `nn_module`.
    Returns a distribution (callable) over `nn.Module`\s, which upon calling
    returns a sampled `nn.Module`.

    :param name: name of pyro module
    :type name: str
    :param nn_module: the module to be registered with pyro
    :type nn_module: torch.nn.Module
    :param prior: pyro distribution, stochastic function, or python dict with parameter names
                  as keys and respective distributions/stochastic functions as values.
    :returns: a callable which returns a sampled module
    """
    warnings.warn(
        "The `random_module` primitive is deprecated, and will be removed "
        "in a future release. Use `pyro.nn.Module` to create Bayesian "
        "modules from `torch.nn.Module` instances.",
        FutureWarning,
    )

    assert hasattr(nn_module, "parameters"), "Module is not a NN module."
    # register params in param store
    lifted_fn = poutine.lift(module, prior=prior)

    def _fn():
        nn_copy = copy.deepcopy(nn_module)
        # update_module_params must be True or the lifted module will not update local params
        return lifted_fn(name, nn_copy, update_module_params=True, *args, **kwargs)

    return _fn


@effectful(type="barrier")
def barrier(data: torch.Tensor) -> torch.Tensor:
    """
    EXPERIMENTAL Ensures all values in ``data`` are ground, rather than lazy
    funsor values. This is useful in combination with
    :func:`pyro.poutine.collapse`.
    """
    return data


def enable_validation(is_validate: bool = True) -> None:
    """
    Enable or disable validation checks in Pyro. Validation checks provide
    useful warnings and errors, e.g. NaN checks, validating distribution
    arguments and support values, detecting incorrect use of ELBO and MCMC.
    Since some of these checks may be expensive, you may want to disable
    validation of mature models to speed up inference.

    The default behavior mimics Python's ``assert`` statement: validation is on
    by default, but is disabled if Python is run in optimized mode (via
    ``python -O``). Equivalently, the default behavior depends on Python's
    global ``__debug__`` value via ``pyro.enable_validation(__debug__)``.

    Validation is temporarily disabled during jit compilation, for all
    inference algorithms that support the PyTorch jit. We recommend developing
    models with non-jitted inference algorithms to ease debugging, then
    optionally moving to jitted inference once a model is correct.

    :param bool is_validate: (optional; defaults to True) whether to
        enable validation checks.
    """
    dist.enable_validation(is_validate)
    infer.enable_validation(is_validate)
    poutine.enable_validation(is_validate)


@contextmanager
def validation_enabled(is_validate: bool = True) -> Iterator[None]:
    """
    Context manager that is useful when temporarily enabling/disabling
    validation checks.

    :param bool is_validate: (optional; defaults to True) temporary
        validation check override.
    """
    infer_validation_status = infer.is_validation_enabled()
    distribution_validation_status = dist.is_validation_enabled()
    poutine_validation_status = poutine.is_validation_enabled()
    try:
        enable_validation(is_validate)
        yield
    finally:
        dist.enable_validation(distribution_validation_status)
        infer.enable_validation(infer_validation_status)
        poutine.enable_validation(poutine_validation_status)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import pyro
import pyro.distributions as dist

from .reparam import Reparam


class StudentTReparam(Reparam):
    """
    Auxiliary variable reparameterizer for
    :class:`~pyro.distributions.StudentT` random variables.

    This is useful in combination with
    :class:`~pyro.infer.reparam.hmm.LinearHMMReparam` because it allows
    StudentT processes to be treated as conditionally Gaussian processes,
    permitting cheap inference via :class:`~pyro.distributions.GaussianHMM` .

    This reparameterizes a :class:`~pyro.distributions.StudentT` by introducing
    an auxiliary :class:`~pyro.distributions.Gamma` variable conditioned on
    which the result is :class:`~pyro.distributions.Normal` .
    """

    def apply(self, msg):
        name = msg["name"]
        fn = msg["fn"]
        value = msg["value"]
        is_observed = msg["is_observed"]

        fn, event_dim = self._unwrap(fn)
        assert isinstance(fn, dist.StudentT)

        # Draw a sample that depends only on df.
        half_df = fn.df * 0.5
        gamma = pyro.sample(
            "{}_gamma".format(name), self._wrap(dist.Gamma(half_df, half_df), event_dim)
        )

        # Construct a scaled Normal.
        loc = fn.loc
        scale = fn.scale * gamma.rsqrt()
        new_fn = self._wrap(dist.Normal(loc, scale), event_dim)
        return {"fn": new_fn, "value": value, "is_observed": is_observed}



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

"""
Poutine is a library of composable effect handlers for recording and modifying the
behavior of Pyro programs. These lower-level ingredients simplify the implementation
of new inference algorithms and behavior.

Handlers can be used as higher-order functions, decorators, or context managers
to modify the behavior of functions or blocks of code:

For example, consider the following Pyro program:

    >>> def model(x):
    ...     s = pyro.param("s", torch.tensor(0.5))
    ...     z = pyro.sample("z", dist.Normal(x, s))
    ...     return z ** 2

We can mark sample sites as observed using ``condition``,
which returns a callable with the same input and output signatures as ``model``:

    >>> conditioned_model = poutine.condition(model, data={"z": 1.0})

We can also use handlers as decorators:

    >>> @pyro.condition(data={"z": 1.0})
    ... def model(x):
    ...     s = pyro.param("s", torch.tensor(0.5))
    ...     z = pyro.sample("z", dist.Normal(x, s))
    ...     return z ** 2

Or as context managers:

    >>> with pyro.condition(data={"z": 1.0}):
    ...     s = pyro.param("s", torch.tensor(0.5))
    ...     z = pyro.sample("z", dist.Normal(0., s))
    ...     y = z ** 2

Handlers compose freely:

    >>> conditioned_model = poutine.condition(model, data={"z": 1.0})
    >>> traced_model = poutine.trace(conditioned_model)

Many inference algorithms or algorithmic components can be implemented
in just a few lines of code::

    guide_tr = poutine.trace(guide).get_trace(...)
    model_tr = poutine.trace(poutine.replay(conditioned_model, trace=guide_tr)).get_trace(...)
    monte_carlo_elbo = model_tr.log_prob_sum() - guide_tr.log_prob_sum()
"""

import collections
import functools
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import ParamSpec

from pyro.poutine import util
from pyro.poutine.block_messenger import BlockMessenger
from pyro.poutine.broadcast_messenger import BroadcastMessenger
from pyro.poutine.collapse_messenger import CollapseMessenger
from pyro.poutine.condition_messenger import ConditionMessenger
from pyro.poutine.do_messenger import DoMessenger
from pyro.poutine.enum_messenger import EnumMessenger
from pyro.poutine.escape_messenger import EscapeMessenger
from pyro.poutine.infer_config_messenger import InferConfigMessenger
from pyro.poutine.lift_messenger import LiftMessenger
from pyro.poutine.markov_messenger import MarkovMessenger
from pyro.poutine.mask_messenger import MaskMessenger
from pyro.poutine.reparam_messenger import ReparamMessenger
from pyro.poutine.replay_messenger import ReplayMessenger
from pyro.poutine.runtime import NonlocalExit
from pyro.poutine.scale_messenger import ScaleMessenger
from pyro.poutine.seed_messenger import SeedMessenger
from pyro.poutine.substitute_messenger import SubstituteMessenger
from pyro.poutine.trace_messenger import TraceMessenger
from pyro.poutine.uncondition_messenger import UnconditionMessenger

if TYPE_CHECKING:
    import numbers

    import torch

    from pyro.distributions.distribution import Distribution
    from pyro.infer.reparam.reparam import Reparam
    from pyro.poutine.runtime import InferDict, Message
    from pyro.poutine.trace_struct import Trace

_P = ParamSpec("_P")
_T = TypeVar("_T")

############################################
# Begin primitive operations
############################################


def _make_handler(msngr_cls, module=None):
    def handler_decorator(func):
        @functools.wraps(func)
        def handler(fn=None, *args, **kwargs):
            if fn is not None and not (
                callable(fn) or isinstance(fn, collections.abc.Iterable)
            ):
                raise ValueError(
                    f"{fn} is not callable, did you mean to pass it as a keyword arg?"
                )
            msngr = msngr_cls(*args, **kwargs)
            return (
                functools.update_wrapper(msngr(fn), fn, updated=())
                if fn is not None
                else msngr
            )

        handler.__doc__ = (
            """Convenient wrapper of :class:`~pyro.poutine.{}.{}` \n\n""".format(
                func.__name__ + "_messenger", msngr_cls.__name__
            )
            + (msngr_cls.__doc__ if msngr_cls.__doc__ else "")
        )
        if module is not None:
            handler.__module__ = module
        return handler

    return handler_decorator


@overload
def block(
    fn: None = ...,
    hide_fn: Optional[Callable[["Message"], Optional[bool]]] = None,
    expose_fn: Optional[Callable[["Message"], Optional[bool]]] = None,
    hide_all: bool = True,
    expose_all: bool = False,
    hide: Optional[List[str]] = None,
    expose: Optional[List[str]] = None,
    hide_types: Optional[List[str]] = None,
    expose_types: Optional[List[str]] = None,
) -> BlockMessenger: ...


@overload
def block(
    fn: Callable[_P, _T] = ...,
    hide_fn: Optional[Callable[["Message"], Optional[bool]]] = None,
    expose_fn: Optional[Callable[["Message"], Optional[bool]]] = None,
    hide_all: bool = True,
    expose_all: bool = False,
    hide: Optional[List[str]] = None,
    expose: Optional[List[str]] = None,
    hide_types: Optional[List[str]] = None,
    expose_types: Optional[List[str]] = None,
) -> Callable[_P, _T]: ...


@_make_handler(BlockMessenger)
def block(  # type: ignore[empty-body]
    fn: Optional[Callable[_P, _T]] = None,
    hide_fn: Optional[Callable[["Message"], Optional[bool]]] = None,
    expose_fn: Optional[Callable[["Message"], Optional[bool]]] = None,
    hide_all: bool = True,
    expose_all: bool = False,
    hide: Optional[List[str]] = None,
    expose: Optional[List[str]] = None,
    hide_types: Optional[List[str]] = None,
    expose_types: Optional[List[str]] = None,
) -> Union[BlockMessenger, Callable[_P, _T]]: ...


@overload
def broadcast(
    fn: None = ...,
) -> BroadcastMessenger: ...


@overload
def broadcast(
    fn: Callable[_P, _T] = ...,
) -> Callable[_P, _T]: ...


@_make_handler(BroadcastMessenger)
def broadcast(  # type: ignore[empty-body]
    fn: Optional[Callable[_P, _T]] = None,
) -> Union[BroadcastMessenger, Callable[_P, _T]]: ...


@overload
def collapse(
    fn: None = ...,
    *args: Any,
    **kwargs: Any,
) -> CollapseMessenger: ...


@overload
def collapse(
    fn: Callable[_P, _T] = ...,
    *args: Any,
    **kwargs: Any,
) -> Callable[_P, _T]: ...


@_make_handler(CollapseMessenger)
def collapse(  # type: ignore[empty-body]
    fn: Optional[Callable[_P, _T]] = None,
    *args: Any,
    **kwargs: Any,
) -> Union[CollapseMessenger, Callable[_P, _T]]: ...


@overload
def condition(
    data: Union[Dict[str, "torch.Tensor"], "Trace"],
) -> ConditionMessenger: ...


@overload
def condition(
    fn: Callable[_P, _T],
    data: Union[Dict[str, "torch.Tensor"], "Trace"],
) -> Callable[_P, _T]: ...


@_make_handler(ConditionMessenger)
def condition(  # type: ignore[empty-body]
    fn: Callable[_P, _T],
    data: Union[Dict[str, "torch.Tensor"], "Trace"],
) -> Union[ConditionMessenger, Callable[_P, _T]]: ...


@overload
def do(
    data: Dict[str, Union["torch.Tensor", "numbers.Number"]],
) -> DoMessenger: ...


@overload
def do(
    fn: Callable[_P, _T],
    data: Dict[str, Union["torch.Tensor", "numbers.Number"]],
) -> Callable[_P, _T]: ...


@_make_handler(DoMessenger)
def do(  # type: ignore[empty-body]
    fn: Callable[_P, _T],
    data: Dict[str, Union["torch.Tensor", "numbers.Number"]],
) -> Union[DoMessenger, Callable[_P, _T]]: ...


@overload
def enum(
    fn: None = ...,
    first_available_dim: Optional[int] = None,
) -> EnumMessenger: ...


@overload
def enum(
    fn: Callable[_P, _T] = ...,
    first_available_dim: Optional[int] = None,
) -> Callable[_P, _T]: ...


@_make_handler(EnumMessenger)
def enum(  # type: ignore[empty-body]
    fn: Optional[Callable[_P, _T]] = None,
    first_available_dim: Optional[int] = None,
) -> Union[EnumMessenger, Callable[_P, _T]]: ...


@overload
def escape(
    escape_fn: Callable[["Message"], bool],
) -> EscapeMessenger: ...


@overload
def escape(
    fn: Callable[_P, _T],
    escape_fn: Callable[["Message"], bool],
) -> Callable[_P, _T]: ...


@_make_handler(EscapeMessenger)
def escape(  # type: ignore[empty-body]
    fn: Callable[_P, _T],
    escape_fn: Callable[["Message"], bool],
) -> Union[EscapeMessenger, Callable[_P, _T]]: ...


@overload
def infer_config(
    config_fn: Callable[["Message"], "InferDict"],
) -> InferConfigMessenger: ...


@overload
def infer_config(
    fn: Callable[_P, _T],
    config_fn: Callable[["Message"], "InferDict"],
) -> Callable[_P, _T]: ...


@_make_handler(InferConfigMessenger)
def infer_config(  # type: ignore[empty-body]
    fn: Callable[_P, _T],
    config_fn: Callable[["Message"], "InferDict"],
) -> Union[InferConfigMessenger, Callable[_P, _T]]: ...


@overload
def lift(
    prior: Union[Callable, "Distribution", Dict[str, Union["Distribution", Callable]]],
) -> LiftMessenger: ...


@overload
def lift(
    fn: Callable[_P, _T],
    prior: Union[Callable, "Distribution", Dict[str, Union["Distribution", Callable]]],
) -> Callable[_P, _T]: ...


@_make_handler(LiftMessenger)
def lift(  # type: ignore[empty-body]
    fn: Callable[_P, _T],
    prior: Union[Callable, "Distribution", Dict[str, Union["Distribution", Callable]]],
) -> Union[LiftMessenger, Callable[_P, _T]]: ...


@overload
def mask(
    mask: Union[bool, "torch.BoolTensor"],
) -> MaskMessenger: ...


@overload
def mask(
    fn: Callable[_P, _T],
    mask: Union[bool, "torch.BoolTensor"],
) -> Callable[_P, _T]: ...


@_make_handler(MaskMessenger)
def mask(  # type: ignore[empty-body]
    fn: Callable[_P, _T],
    mask: Union[bool, "torch.BoolTensor"],
) -> Union[MaskMessenger, Callable[_P, _T]]: ...


@overload
def reparam(
    config: Union[Dict[str, "Reparam"], Callable[["Message"], Optional["Reparam"]]],
) -> ReparamMessenger: ...


@overload
def reparam(
    fn: Callable[_P, _T],
    config: Union[Dict[str, "Reparam"], Callable[["Message"], Optional["Reparam"]]],
) -> Callable[_P, _T]: ...


@_make_handler(ReparamMessenger)
def reparam(  # type: ignore[empty-body]
    fn: Callable[_P, _T],
    config: Union[Dict[str, "Reparam"], Callable[["Message"], Optional["Reparam"]]],
) -> Union[ReparamMessenger, Callable[_P, _T]]: ...


@overload
def replay(
    fn: None = ...,
    trace: Optional["Trace"] = None,
    params: Optional[Dict[str, "torch.Tensor"]] = None,
) -> ReplayMessenger: ...


@overload
def replay(
    fn: Callable[_P, _T] = ...,
    trace: Optional["Trace"] = None,
    params: Optional[Dict[str, "torch.Tensor"]] = None,
) -> Callable[_P, _T]: ...


@_make_handler(ReplayMessenger)
def replay(  # type: ignore[empty-body]
    fn: Optional[Callable[_P, _T]] = None,
    trace: Optional["Trace"] = None,
    params: Optional[Dict[str, "torch.Tensor"]] = None,
) -> Union[ReplayMessenger, Callable[_P, _T]]: ...


@overload
def scale(
    scale: Union[float, "torch.Tensor"],
) -> ScaleMessenger: ...


@overload
def scale(
    fn: Callable[_P, _T],
    scale: Union[float, "torch.Tensor"],
) -> Callable[_P, _T]: ...


@_make_handler(ScaleMessenger)
def scale(  # type: ignore[empty-body]
    fn: Callable[_P, _T],
    scale: Union[float, "torch.Tensor"],
) -> Union[ScaleMessenger, Callable[_P, _T]]: ...


@overload
def seed(
    rng_seed: int,
) -> SeedMessenger: ...


@overload
def seed(
    fn: Callable[_P, _T],
    rng_seed: int,
) -> Callable[_P, _T]: ...


@_make_handler(SeedMessenger)
def seed(  # type: ignore[empty-body]
    fn: Callable[_P, _T],
    rng_seed: int,
) -> Union[SeedMessenger, Callable[_P, _T]]: ...


@overload
def substitute(
    data: Dict[str, "torch.Tensor"],
) -> SubstituteMessenger: ...


@overload
def substitute(
    fn: Callable[_P, _T],
    data: Dict[str, "torch.Tensor"],
) -> Callable[_P, _T]: ...


@_make_handler(SubstituteMessenger)
def substitute(  # type: ignore[empty-body]
    fn: Callable[_P, _T],
    data: Dict[str, "torch.Tensor"],
) -> Union[SubstituteMessenger, Callable[_P, _T]]: ...


@overload
def trace(
    fn: None = ...,
    graph_type: Optional[Literal["flat", "dense"]] = None,
    param_only: Optional[bool] = None,
) -> TraceMessenger: ...


@overload
def trace(
    fn: Callable[_P, _T] = ...,
    graph_type: Optional[Literal["flat", "dense"]] = None,
    param_only: Optional[bool] = None,
) -> Callable[_P, _T]: ...


@_make_handler(TraceMessenger)
def trace(  # type: ignore[empty-body]
    fn: Optional[Callable[_P, _T]] = None,
    graph_type: Optional[Literal["flat", "dense"]] = None,
    param_only: Optional[bool] = None,
) -> Union[TraceMessenger, Callable[_P, _T]]: ...


@overload
def uncondition(
    fn: None = ...,
) -> UnconditionMessenger: ...


@overload
def uncondition(
    fn: Callable[_P, _T] = ...,
) -> Callable[_P, _T]: ...


@_make_handler(UnconditionMessenger)
def uncondition(  # type: ignore[empty-body]
    fn: Optional[Callable[_P, _T]] = None,
) -> Union[UnconditionMessenger, Callable[_P, _T]]: ...


#########################################
# Begin composite operations
#########################################


def queue(
    fn=None,
    queue=None,
    max_tries=None,
    extend_fn=None,
    escape_fn=None,
    num_samples=None,
):
    """
    Used in sequential enumeration over discrete variables.

    Given a stochastic function and a queue,
    return a return value from a complete trace in the queue.

    :param fn: a stochastic function (callable containing Pyro primitive calls)
    :param queue: a queue data structure like multiprocessing.Queue to hold partial traces
    :param max_tries: maximum number of attempts to compute a single complete trace
    :param extend_fn: function (possibly stochastic) that takes a partial trace and a site,
        and returns a list of extended traces
    :param escape_fn: function (possibly stochastic) that takes a partial trace and a site,
        and returns a boolean value to decide whether to exit
    :param num_samples: optional number of extended traces for extend_fn to return
    :returns: stochastic function decorated with poutine logic
    """

    if max_tries is None:
        max_tries = int(1e6)

    if extend_fn is None:
        extend_fn = util.enum_extend

    if escape_fn is None:
        escape_fn = util.discrete_escape

    if num_samples is None:
        num_samples = -1

    def wrapper(wrapped):
        def _fn(*args, **kwargs):
            for i in range(max_tries):
                assert (
                    not queue.empty()
                ), "trying to get() from an empty queue will deadlock"

                next_trace = queue.get()
                try:
                    ftr = trace(
                        escape(
                            replay(wrapped, trace=next_trace),  # noqa: F821
                            escape_fn=functools.partial(escape_fn, next_trace),
                        )
                    )
                    return ftr(*args, **kwargs)
                except NonlocalExit as site_container:
                    site_container.reset_stack()
                    for tr in extend_fn(
                        ftr.trace.copy(), site_container.site, num_samples=num_samples
                    ):
                        queue.put(tr)

            raise ValueError("max tries ({}) exceeded".format(str(max_tries)))

        return _fn

    return wrapper(fn) if fn is not None else wrapper


@overload
def markov(
    fn: None = ...,
    history: int = 1,
    keep: bool = False,
    dim: Optional[int] = None,
    name: Optional[str] = None,
) -> MarkovMessenger: ...


@overload
def markov(
    fn: Iterable[int] = ...,
    history: int = 1,
    keep: bool = False,
    dim: Optional[int] = None,
    name: Optional[str] = None,
) -> MarkovMessenger: ...


@overload
def markov(
    fn: Callable[_P, _T] = ...,
    history: int = 1,
    keep: bool = False,
    dim: Optional[int] = None,
    name: Optional[str] = None,
) -> Callable[_P, _T]: ...


def markov(
    fn: Optional[Union[Iterable[int], Callable[_P, _T]]] = None,
    history: int = 1,
    keep: bool = False,
    dim: Optional[int] = None,
    name: Optional[str] = None,
) -> Union[MarkovMessenger, Callable[_P, _T]]:
    """
    Markov dependency declaration.

    This can be used in a variety of ways:

        - as a context manager
        - as a decorator for recursive functions
        - as an iterator for markov chains

    :param int history: The number of previous contexts visible from the
        current context. Defaults to 1. If zero, this is similar to
        :class:`pyro.plate`.
    :param bool keep: If true, frames are replayable. This is important
        when branching: if ``keep=True``, neighboring branches at the same
        level can depend on each other; if ``keep=False``, neighboring branches
        are independent (conditioned on their share"
    :param int dim: An optional dimension to use for this independence index.
        Interface stub, behavior not yet implemented.
    :param str name: An optional unique name to help inference algorithms match
        :func:`pyro.markov` sites between models and guides.
        Interface stub, behavior not yet implemented.
    """
    if fn is None:
        # Used as a decorator with bound args
        return MarkovMessenger(history=history, keep=keep, dim=dim, name=name)
    if not callable(fn):
        # Used as a generator
        return MarkovMessenger(
            history=history, keep=keep, dim=dim, name=name
        ).generator(iterable=fn)
    # Used as a decorator with bound args
    return MarkovMessenger(history=history, keep=keep, dim=dim, name=name)(fn)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import warnings
from abc import ABC
from typing import Callable, Optional

import torch
from typing_extensions import TypedDict


class ReparamMessage(TypedDict):
    name: str
    fn: Callable
    value: Optional[torch.Tensor]
    is_observed: Optional[bool]


class ReparamResult(TypedDict):
    fn: Callable
    value: Optional[torch.Tensor]
    is_observed: bool


class Reparam(ABC):
    """
    Abstract base class for reparameterizers.

    Derived classes should implement :meth:`apply`.
    """

    # @abstractmethod  # Not abstract, for backwards compatibility.
    def apply(self, msg: ReparamMessage) -> ReparamResult:
        """
        Abstract method to apply reparameterizer.

        :param dict name: A simplified Pyro message with fields:
            - ``name: str`` the sample site's name
            - ``fn: Callable`` a distribution
            - ``value: Optional[torch.Tensor]`` an observed or initial value
            - ``is_observed: bool`` whether ``value`` is an observation
        :returns: A simplified Pyro message with fields ``fn``, ``value``, and
            ``is_observed``.
        :rtype: dict
        """

        # This default is provided for backwards compatibility only.
        # New subclasses should define .apply() and omit .__call__().
        warnings.warn(
            "Reparam.__call__() is deprecated in favor of .apply(); "
            "new subclasses should implement .apply().",
            DeprecationWarning,
        )
        new_fn, value = self(msg["name"], msg["fn"], msg["value"])
        is_observed = msg["value"] is None and value is not None
        return {"fn": new_fn, "value": value, "is_observed": is_observed}

    def __call__(self, name, fn, obs):
        """
        DEPRECATED.
        Subclasses should implement :meth:`apply` instead.
        This will be removed in a future release.
        """
        raise NotImplementedError

    def _unwrap(self, fn):
        """
        Unwrap Independent distributions.
        """
        event_dim = fn.event_dim
        while isinstance(fn, torch.distributions.Independent):
            fn = fn.base_dist
        return fn, event_dim

    def _wrap(self, fn, event_dim):
        """
        Wrap in Independent distributions.
        """
        if fn.event_dim < event_dim:
            fn = fn.to_event(event_dim - fn.event_dim)
        assert fn.event_dim == event_dim
        return fn



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import functools
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    overload,
)

import torch
from typing_extensions import Literal, ParamSpec, TypedDict

from pyro.params.param_store import (  # noqa: F401
    _MODULE_NAMESPACE_DIVIDER,
    ParamStoreDict,
)

_P = ParamSpec("_P")
_T = TypeVar("_T")

if TYPE_CHECKING:
    from collections import Counter

    from pyro.distributions.score_parts import ScoreParts
    from pyro.distributions.torch_distribution import TorchDistributionMixin
    from pyro.poutine.indep_messenger import CondIndepStackFrame
    from pyro.poutine.messenger import Messenger

# the global pyro stack
_PYRO_STACK: List["Messenger"] = []

# the global ParamStore
_PYRO_PARAM_STORE = ParamStoreDict()


class InferDict(TypedDict, total=False):
    """
    A dictionary that contains information about inference.

    This can be used to configure per-site inference strategies, e.g.::

        pyro.sample(
            "x",
            dist.Bernoulli(0.5),
            infer={"enumerate": "parallel"},
        )

    Keys:
        enumerate (str):
            If one of the strings "sequential" or "parallel", enables
            enumeration. Parallel enumeration is generally faster but requires
            broadcasting-safe operations and static structure.
        expand (bool):
            Whether to expand the distribution during enumeration. Defaults to
            False if missing.
        is_auxiliary (bool):
            Whether the sample site is auxiliary, e.g. for use in guides that
            deterministically transform auxiliary variables. Defaults to False
            if missing.
        is_observed (bool):
            Whether the sample site is observed (i.e. not latent). Defaults to
            False if missing.
        num_samples (int):
            The number of samples to draw. Defaults to 1 if missing.
        obs (optional torch.Tensor):
            The observed value, or None for latent variables. Defaults to None
            if missing.
        prior (optional torch.distributions.Distribution):
            (internal) For use in GuideMessenger to store the model's prior
            distribution (conditioned on upstream sites).
        tmc (str):
            Whether to use the diagonal or mixture approximation for Tensor
            Monte Carlo in TraceTMC_ELBO.
        was_observed (bool):
            (internal) Whether the sample site was originally observed, in the
            context of inference via Reweighted Wake Sleep or Compiled
            Sequential Importance Sampling.
    """

    enumerate: Literal["sequential", "parallel"]
    expand: bool
    is_auxiliary: bool
    is_observed: bool
    num_samples: int
    obs: Optional[torch.Tensor]
    prior: "TorchDistributionMixin"
    tmc: Literal["diagonal", "mixture"]
    was_observed: bool
    _deterministic: bool
    _dim_to_symbol: Dict[int, str]
    _do_not_trace: bool
    _enumerate_symbol: str
    _markov_scope: "Counter"
    _enumerate_dim: int
    _dim_to_id: Dict[int, int]
    _markov_depth: int


class Message(TypedDict, Generic[_P, _T], total=False):
    """
    Pyro's internal message type for effect handling.

    Messages are stored in trace objects, e.g.::

        trace.nodes["my_site_name"]  # This is a Message.

    Keys:
        type (str):
            The message type, typically one of the strings "sample", "param",
            "plate", or "markov", but possibly custom.
        name (str):
            The site name, typically naming a sample or parameter.
        fn (callable):
            The distribution or function used to generate the sample.
        is_observed (bool):
            A flag to indicate whether the value is observed.
        args (tuple):
            Positional arguments to the distribution or function.
        kwargs (dict):
            Keyword arguments to the distribution or function.
        value (torch.Tensor):
            The value of the sample (either observed or sampled).
        scale (torch.Tensor):
            A scaling factor for the log probability.
        mask (bool torch.Tensor):
            A bool or tensor to mask the log probability.
        cond_indep_stack (tuple):
            The site's local stack of conditional independence metadata.
            Immutable.
        done (bool):
            A flag to indicate whether the message has been handled.
        stop (bool):
            A flag to stop further processing of the message.
        continuation (callable):
            A function to call after processing the message.
        infer (optional InferDict):
            A dictionary of inference parameters.
        obs (torch.Tensor):
            The observed value.
        log_prob (torch.Tensor):
            The log probability of the sample.
        log_prob_sum (torch.Tensor):
            The sum of the log probability.
        unscaled_log_prob (torch.Tensor):
            The unscaled log probability.
        score_parts (pyro.distributions.ScoreParts):
            A collection of score parts.
        packed (Message):
            A packed message, used during enumeration.
    """

    type: str
    name: Optional[str]
    fn: Callable[_P, _T]
    is_observed: bool
    args: Tuple
    kwargs: Dict
    value: Optional[_T]
    scale: Union[torch.Tensor, float]
    mask: Union[bool, torch.Tensor, None]
    cond_indep_stack: Tuple["CondIndepStackFrame", ...]
    done: bool
    stop: bool
    continuation: Optional[Callable[["Message"], None]]
    infer: Optional[InferDict]
    obs: Optional[torch.Tensor]
    log_prob: torch.Tensor
    log_prob_sum: torch.Tensor
    unscaled_log_prob: torch.Tensor
    score_parts: "ScoreParts"
    packed: "Message"
    _intervener_id: Optional[str]


class _DimAllocator:
    """
    Dimension allocator for internal use by :class:`plate`.
    There is a single global instance.

    Note that dimensions are indexed from the right, e.g. -1, -2.
    """

    def __init__(self) -> None:
        # in reverse orientation of log_prob.shape
        self._stack: List[Optional[str]] = []

    def allocate(self, name: str, dim: Optional[int]) -> int:
        """
        Allocate a dimension to an :class:`plate` with given name.
        Dim should be either None for automatic allocation or a negative
        integer for manual allocation.
        """
        if name in self._stack:
            raise ValueError(f"duplicate plate '{name}'")
        if dim is None:
            # Automatically designate the rightmost available dim for allocation.
            dim = -1
            while -dim <= len(self._stack) and self._stack[-1 - dim] is not None:
                dim -= 1
        elif dim >= 0:
            raise ValueError(f"Expected dim < 0 to index from the right, actual {dim}")

        # Allocate the requested dimension.
        while dim < -len(self._stack):
            self._stack.append(None)
        if self._stack[-1 - dim] is not None:
            raise ValueError(
                "\n".join(
                    [
                        'at plates "{}" and "{}", collide at dim={}'.format(
                            name, self._stack[-1 - dim], dim
                        ),
                        "\nTry moving the dim of one plate to the left, e.g. dim={}".format(
                            dim - 1
                        ),
                    ]
                )
            )
        self._stack[-1 - dim] = name
        return dim

    def free(self, name: str, dim: int) -> None:
        """
        Free a dimension.
        """
        free_idx = -1 - dim  # stack index to free
        assert self._stack[free_idx] == name
        self._stack[free_idx] = None
        while self._stack and self._stack[-1] is None:
            self._stack.pop()


# Handles placement of plate dimensions
_DIM_ALLOCATOR = _DimAllocator()


class _EnumAllocator:
    """
    Dimension allocator for internal use by :func:`~pyro.poutine.markov`.
    There is a single global instance.

    Note that dimensions are indexed from the right, e.g. -1, -2.
    Note that ids are simply nonnegative integers here.
    """

    def set_first_available_dim(self, first_available_dim: int) -> None:
        """
        Set the first available dim, which should be to the left of all
        :class:`plate` dimensions, e.g. ``-1 - max_plate_nesting``. This should
        be called once per program. In SVI this should be called only once per
        (guide,model) pair.
        """
        assert first_available_dim < 0, first_available_dim
        self.next_available_dim = first_available_dim
        self.next_available_id = 0
        self.dim_to_id: Dict[int, int] = {}  # only the global ids

    def allocate(self, scope_dims: Optional[Set[int]] = None) -> Tuple[int, int]:
        """
        Allocate a new recyclable dim and a unique id.

        If ``scope_dims`` is None, this allocates a global enumeration dim
        that will never be recycled. If ``scope_dims`` is specified, this
        allocates a local enumeration dim that can be reused by at any other
        local site whose scope excludes this site.

        :param set scope_dims: An optional set of (negative integer)
            local enumeration dims to avoid when allocating this dim.
        :return: A pair ``(dim, id)``, where ``dim`` is a negative integer
            and ``id`` is a nonnegative integer.
        :rtype: tuple
        """
        id_ = self.next_available_id
        self.next_available_id += 1

        dim = self.next_available_dim
        if dim == -float("inf"):
            raise ValueError(
                "max_plate_nesting must be set to a finite value for parallel enumeration"
            )
        if scope_dims is None:
            # allocate a new global dimension
            self.next_available_dim -= 1
            self.dim_to_id[dim] = id_
        else:
            # allocate a new local dimension
            while dim in scope_dims:
                dim -= 1

        return dim, id_


# Handles placement of enumeration dimensions
_ENUM_ALLOCATOR = _EnumAllocator()


class NonlocalExit(Exception):
    """
    Exception for exiting nonlocally from poutine execution.

    Used by poutine.EscapeMessenger to return site information.
    """

    def __init__(self, site: Message, *args, **kwargs) -> None:
        """
        :param site: message at a pyro site constructor.
            Just stores the input site.
        """
        super().__init__(*args, **kwargs)
        self.site = site

    def reset_stack(self) -> None:
        """
        Reset the state of the frames remaining in the stack.
        Necessary for multiple re-executions in poutine.queue.
        """
        from pyro.poutine.block_messenger import BlockMessenger

        for frame in reversed(_PYRO_STACK):
            frame._reset()
            if isinstance(frame, BlockMessenger) and frame.hide_fn(self.site):
                break


def default_process_message(msg: Message) -> None:
    """
    Default method for processing messages in inference.

    :param msg: a message to be processed
    :returns: None
    """
    if msg["done"] or msg["is_observed"] or msg["value"] is not None:
        msg["done"] = True
        return

    msg["value"] = msg["fn"](*msg["args"], **msg["kwargs"])

    # after fn has been called, update msg to prevent it from being called again.
    msg["done"] = True


def apply_stack(initial_msg: Message) -> None:
    """
    Execute the effect stack at a single site according to the following scheme:

        1. For each ``Messenger`` in the stack from bottom to top,
           execute ``Messenger._process_message`` with the message;
           if the message field "stop" is True, stop;
           otherwise, continue
        2. Apply default behavior (``default_process_message``) to finish remaining site execution
        3. For each ``Messenger`` in the stack from top to bottom,
           execute ``_postprocess_message`` to update the message and internal messenger state with the site results
        4. If the message field "continuation" is not ``None``, call it with the message

    :param dict initial_msg: the starting version of the trace site
    :returns: ``None``
    """
    stack = _PYRO_STACK
    # TODO check at runtime if stack is valid

    # msg is used to pass information up and down the stack
    msg = initial_msg

    pointer = 0
    # go until time to stop?
    for frame in reversed(stack):
        pointer = pointer + 1

        frame._process_message(msg)

        if msg["stop"]:
            break

    default_process_message(msg)

    for frame in stack[-pointer:]:
        frame._postprocess_message(msg)

    cont = msg["continuation"]
    if cont is not None:
        cont(msg)


def am_i_wrapped() -> bool:
    """
    Checks whether the current computation is wrapped in a poutine.
    :returns: bool
    """
    return len(_PYRO_STACK) > 0


@overload
def effectful(
    fn: None = ..., type: Optional[str] = ...
) -> Callable[[Callable[_P, _T]], Callable[..., Optional[_T]]]: ...


@overload
def effectful(
    fn: Callable[_P, _T] = ..., type: Optional[str] = ...
) -> Callable[..., Optional[_T]]: ...


def effectful(
    fn: Optional[Callable[_P, _T]] = None, type: Optional[str] = None
) -> Callable:
    """
    :param fn: function or callable that performs an effectful computation
    :param str type: the type label of the operation, e.g. `"sample"`

    Wrapper for calling :func:`~pyro.poutine.runtime.apply_stack` to apply any active effects.
    """
    if fn is None:
        return functools.partial(effectful, type=type)

    if getattr(fn, "_is_effectful", None):
        return fn

    assert type is not None, f"must provide a type label for operation {fn}"
    assert type != "message", "cannot use 'message' as keyword"

    @functools.wraps(fn)
    def _fn(
        *args: _P.args,
        name: Optional[str] = None,
        infer: Optional[InferDict] = None,
        obs: Optional[_T] = None,
        **kwargs: _P.kwargs,
    ) -> Optional[_T]:
        is_observed = obs is not None

        if not am_i_wrapped():
            return fn(*args, **kwargs)
        else:
            msg = Message(
                type=type,
                name=name,
                fn=fn,
                is_observed=is_observed,
                args=args,
                kwargs=kwargs,
                value=obs,
                scale=1.0,
                mask=None,
                cond_indep_stack=(),
                done=False,
                stop=False,
                continuation=None,
                infer=infer if infer is not None else {},
            )
            # apply the stack and return its return value
            apply_stack(msg)
            return msg["value"]

    _fn._is_effectful = True  # type: ignore[attr-defined]
    return _fn


def _inspect() -> Message:
    """
    EXPERIMENTAL Inspect the Pyro stack.

    .. warning:: The format of the returned message may change at any time and
        does not guarantee backwards compatibility.

    :returns: A message with all effects applied.
    :rtype: dict
    """
    msg = Message(
        type="inspect",
        name="_pyro_inspect",
        fn=lambda: True,
        is_observed=False,
        args=(),
        kwargs={},
        value=None,
        infer={"_do_not_trace": True},
        scale=1.0,
        mask=None,
        cond_indep_stack=(),
        done=False,
        stop=False,
        continuation=None,
    )
    apply_stack(msg)
    return msg


def get_mask() -> Union[bool, torch.Tensor, None]:
    """
    Records the effects of enclosing ``poutine.mask`` handlers.

    This is useful for avoiding expensive ``pyro.factor()`` computations during
    prediction, when the log density need not be computed, e.g.::

        def model():
            # ...
            if poutine.get_mask() is not False:
                log_density = my_expensive_computation()
                pyro.factor("foo", log_density)
            # ...

    :returns: The mask.
    :rtype: None, bool, or torch.Tensor
    """
    return _inspect()["mask"]


def get_plates() -> Tuple["CondIndepStackFrame", ...]:
    """
    Records the effects of enclosing ``pyro.plate`` contexts.

    :returns: A tuple of
        :class:`pyro.poutine.indep_messenger.CondIndepStackFrame` objects.
    :rtype: tuple
    """
    return _inspect()["cond_indep_stack"]



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import math
import warnings

import torch

import pyro
import pyro.poutine as poutine
from pyro.ops.stats import fit_generalized_pareto

from .abstract_infer import TracePosterior
from .enum import get_importance_trace


class Importance(TracePosterior):
    """
    :param model: probabilistic model defined as a function
    :param guide: guide used for sampling defined as a function
    :param num_samples: number of samples to draw from the guide (default 10)

    This method performs posterior inference by importance sampling
    using the guide as the proposal distribution.
    If no guide is provided, it defaults to proposing from the model's prior.
    """

    def __init__(self, model, guide=None, num_samples=None):
        """
        Constructor. default to num_samples = 10, guide = model
        """
        super().__init__()
        if num_samples is None:
            num_samples = 10
            warnings.warn(
                "num_samples not provided, defaulting to {}".format(num_samples)
            )
        if guide is None:
            # propose from the prior by making a guide from the model by hiding observes
            guide = poutine.block(model, hide_types=["observe"])
        self.num_samples = num_samples
        self.model = model
        self.guide = guide

    def _traces(self, *args, **kwargs):
        """
        Generator of weighted samples from the proposal distribution.
        """
        for i in range(self.num_samples):
            guide_trace = poutine.trace(self.guide).get_trace(*args, **kwargs)
            model_trace = poutine.trace(
                poutine.replay(self.model, trace=guide_trace)
            ).get_trace(*args, **kwargs)
            log_weight = model_trace.log_prob_sum() - guide_trace.log_prob_sum()
            yield (model_trace, log_weight)

    def get_log_normalizer(self):
        """
        Estimator of the normalizing constant of the target distribution.
        (mean of the unnormalized weights)
        """
        # ensure list is not empty
        if self.log_weights:
            log_w = torch.tensor(self.log_weights)
            log_num_samples = torch.log(torch.tensor(self.num_samples * 1.0))
            return torch.logsumexp(log_w - log_num_samples, 0)
        else:
            warnings.warn(
                "The log_weights list is empty, can not compute normalizing constant estimate."
            )

    def get_normalized_weights(self, log_scale=False):
        """
        Compute the normalized importance weights.
        """
        if self.log_weights:
            log_w = torch.tensor(self.log_weights)
            log_w_norm = log_w - torch.logsumexp(log_w, 0)
            return log_w_norm if log_scale else torch.exp(log_w_norm)
        else:
            warnings.warn(
                "The log_weights list is empty. There is nothing to normalize."
            )

    def get_ESS(self):
        """
        Compute (Importance Sampling) Effective Sample Size (ESS).
        """
        if self.log_weights:
            log_w_norm = self.get_normalized_weights(log_scale=True)
            ess = torch.exp(-torch.logsumexp(2 * log_w_norm, 0))
        else:
            warnings.warn(
                "The log_weights list is empty, effective sample size is zero."
            )
            ess = 0
        return ess


def vectorized_importance_weights(model, guide, *args, **kwargs):
    """
    :param model: probabilistic model defined as a function
    :param guide: guide used for sampling defined as a function
    :param num_samples: number of samples to draw from the guide (default 1)
    :param int max_plate_nesting: Bound on max number of nested :func:`pyro.plate` contexts.
    :param bool normalized: set to True to return self-normalized importance weights
    :returns: returns a ``(num_samples,)``-shaped tensor of importance weights
        and the model and guide traces that produced them

    Vectorized computation of importance weights for models with static structure::

        log_weights, model_trace, guide_trace = \\
            vectorized_importance_weights(model, guide, *args,
                                          num_samples=1000,
                                          max_plate_nesting=4,
                                          normalized=False)
    """
    num_samples = kwargs.pop("num_samples", 1)
    max_plate_nesting = kwargs.pop("max_plate_nesting", None)
    normalized = kwargs.pop("normalized", False)

    if max_plate_nesting is None:
        raise ValueError("must provide max_plate_nesting")
    max_plate_nesting += 1

    def vectorize(fn):
        def _fn(*args, **kwargs):
            with pyro.plate(
                "num_particles_vectorized", num_samples, dim=-max_plate_nesting
            ):
                return fn(*args, **kwargs)

        return _fn

    model_trace, guide_trace = get_importance_trace(
        "flat", max_plate_nesting, vectorize(model), vectorize(guide), args, kwargs
    )

    guide_trace.pack_tensors()
    model_trace.pack_tensors(guide_trace.plate_to_symbol)

    if num_samples == 1:
        log_weights = model_trace.log_prob_sum() - guide_trace.log_prob_sum()
    else:
        wd = guide_trace.plate_to_symbol["num_particles_vectorized"]
        log_weights = 0.0
        for site in model_trace.nodes.values():
            if site["type"] != "sample":
                continue
            log_weights += torch.einsum(
                site["packed"]["log_prob"]._pyro_dims + "->" + wd,
                [site["packed"]["log_prob"]],
            )

        for site in guide_trace.nodes.values():
            if site["type"] != "sample":
                continue
            log_weights -= torch.einsum(
                site["packed"]["log_prob"]._pyro_dims + "->" + wd,
                [site["packed"]["log_prob"]],
            )

    if normalized:
        log_weights = log_weights - torch.logsumexp(log_weights)
    return log_weights, model_trace, guide_trace


@torch.no_grad()
def psis_diagnostic(model, guide, *args, **kwargs):
    """
    Computes the Pareto tail index k for a model/guide pair using the technique
    described in [1], which builds on previous work in [2]. If :math:`0 < k < 0.5`
    the guide is a good approximation to the model posterior, in the sense
    described in [1]. If :math:`0.5 \\le k \\le 0.7`, the guide provides a suboptimal
    approximation to the posterior, but may still be useful in practice. If
    :math:`k > 0.7` the guide program provides a poor approximation to the full
    posterior, and caution should be used when using the guide. Note, however,
    that a guide may be a poor fit to the full posterior while still yielding
    reasonable model predictions. If :math:`k < 0.0` the importance weights
    corresponding to the model and guide appear to be bounded from above; this
    would be a bizarre outcome for a guide trained via ELBO maximization. Please
    see [1] for a more complete discussion of how the tail index k should be
    interpreted.

    Please be advised that a large number of samples may be required for an
    accurate estimate of k.

    Note that we assume that the model and guide are both vectorized and have
    static structure. As is canonical in Pyro, the args and kwargs are passed
    to the model and guide.

    References
    [1] 'Yes, but Did It Work?: Evaluating Variational Inference.'
    Yuling Yao, Aki Vehtari, Daniel Simpson, Andrew Gelman
    [2] 'Pareto Smoothed Importance Sampling.'
    Aki Vehtari, Andrew Gelman, Jonah Gabry

    :param callable model: the model program.
    :param callable guide: the guide program.
    :param int num_particles: the total number of times we run the model and guide in
        order to compute the diagnostic. defaults to 1000.
    :param max_simultaneous_particles: the maximum number of simultaneous samples drawn
        from the model and guide. defaults to `num_particles`. `num_particles` must be
        divisible by `max_simultaneous_particles`. compute the diagnostic. defaults to 1000.
    :param int max_plate_nesting: optional bound on max number of nested :func:`pyro.plate`
        contexts in the model/guide. defaults to 7.
    :returns float: the PSIS diagnostic k
    """

    num_particles = kwargs.pop("num_particles", 1000)
    max_simultaneous_particles = kwargs.pop("max_simultaneous_particles", num_particles)
    max_plate_nesting = kwargs.pop("max_plate_nesting", 7)

    if num_particles % max_simultaneous_particles != 0:
        raise ValueError(
            "num_particles must be divisible by max_simultaneous_particles."
        )

    N = num_particles // max_simultaneous_particles
    log_weights = [
        vectorized_importance_weights(
            model,
            guide,
            num_samples=max_simultaneous_particles,
            max_plate_nesting=max_plate_nesting,
            *args,
            **kwargs,
        )[0]
        for _ in range(N)
    ]
    log_weights = torch.cat(log_weights)
    log_weights -= log_weights.max()
    log_weights = torch.sort(log_weights, descending=False)[0]

    cutoff_index = (
        -int(math.ceil(min(0.2 * num_particles, 3.0 * math.sqrt(num_particles)))) - 1
    )
    lw_cutoff = max(math.log(1.0e-15), log_weights[cutoff_index])
    lw_tail = log_weights[log_weights > lw_cutoff]

    if len(lw_tail) < 10:
        warnings.warn(
            "Not enough tail samples to compute PSIS diagnostic; increase num_particles."
        )
        k = float("inf")
    else:
        k, _ = fit_generalized_pareto(lw_tail.exp() - math.exp(lw_cutoff))

    return k



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

from .handlers import (
    block,
    broadcast,
    collapse,
    condition,
    do,
    enum,
    escape,
    infer_config,
    lift,
    markov,
    mask,
    queue,
    reparam,
    replay,
    scale,
    seed,
    substitute,
    trace,
    uncondition,
)
from .messenger import unwrap
from .runtime import NonlocalExit, get_mask
from .trace_struct import Trace
from .util import enable_validation, is_validation_enabled

__all__ = [
    "block",
    "broadcast",
    "collapse",
    "condition",
    "do",
    "enable_validation",
    "enum",
    "escape",
    "get_mask",
    "infer_config",
    "is_validation_enabled",
    "lift",
    "markov",
    "mask",
    "NonlocalExit",
    "replay",
    "reparam",
    "queue",
    "scale",
    "seed",
    "substitute",
    "trace",
    "Trace",
    "uncondition",
    "unwrap",
]
